# -*- coding: utf-8 -*-

"""The database manager for PyBEL.

Under the hood, PyBEL caches namespace and annotation files for quick recall on later use. The user doesn't need to
enable this option, but can specify a database location if they choose.
"""

import logging
import time
from typing import Iterable, List, Mapping, Optional, Set, Tuple

import pandas as pd
import requests
import sqlalchemy
from bel_resources import get_bel_resource
from sqlalchemy import and_, exists, func
from sqlalchemy.orm import aliased
from tqdm.autonotebook import tqdm

from .base_manager import BaseManager, build_engine_session
from .exc import EdgeAddError
from .lookup_manager import LookupManager
from .models import (
    Author,
    Citation,
    Edge,
    Evidence,
    Namespace,
    NamespaceEntry,
    Network,
    Node,
    edge_annotation,
    network_edge,
    network_node,
)
from .query_manager import QueryManager
from .utils import (
    extract_shared_optional,
    extract_shared_required,
    update_insert_values,
)
from ..constants import (
    ANNOTATIONS,
    CITATION,
    CITATION_TYPE_PUBMED,
    EVIDENCE,
    IDENTIFIER,
    METADATA_INSERT_KEYS,
    NAMESPACE,
    RELATION,
    SOURCE_MODIFIER,
    TARGET_MODIFIER,
    UNQUALIFIED_EDGES,
    belns_encodings,
    get_cache_connection,
)
from ..dsl import BaseConcept, BaseEntity
from ..language import Entity
from ..struct.graph import AnnotationsDict, BELGraph
from ..struct.operations import union
from ..typing import EdgeData

__all__ = [
    "Manager",
    "NetworkManager",
]

logger = logging.getLogger(__name__)

DEFAULT_BELNS_ENCODING = "".join(sorted(belns_encodings))

_optional_namespace_entries_mapping = {
    "species": ("Namespace", "SpeciesString"),
    "query_url": ("Namespace", "QueryValueURL"),
    "domain": ("Namespace", "DomainString"),
}


def _get_namespace_insert_values(bel_resource):
    namespace_insert_values = {
        "name": bel_resource["Namespace"]["NameString"],
    }

    namespace_insert_values.update(extract_shared_required(bel_resource, "Namespace"))
    namespace_insert_values.update(extract_shared_optional(bel_resource, "Namespace"))

    update_insert_values(
        bel_resource=bel_resource,
        mapping=_optional_namespace_entries_mapping,
        values=namespace_insert_values,
    )

    return namespace_insert_values


_annotation_mapping = {
    "name": ("Citation", "NameString"),
}


def _get_annotation_insert_values(bel_resource):
    annotation_insert_values = extract_shared_required(bel_resource, "AnnotationDefinition")
    annotation_insert_values.update(extract_shared_optional(bel_resource, "AnnotationDefinition"))
    update_insert_values(
        bel_resource=bel_resource,
        mapping=_annotation_mapping,
        values=annotation_insert_values,
    )
    return annotation_insert_values


def not_resource_cachable(bel_resource):
    """Check if the BEL resource is cacheable.

    :param dict bel_resource: A dictionary returned by :func:`get_bel_resource`.
    """
    return bel_resource["Processing"].get("CacheableFlag") not in {
        "yes",
        "Yes",
        "True",
        "true",
    }


def _clean_bel_namespace_values(bel_resource):
    bel_resource["Values"] = {
        name: (encoding if encoding else DEFAULT_BELNS_ENCODING)
        for name, encoding in bel_resource["Values"].items()
        if name
    }


class NamespaceManager(BaseManager):
    """Manages BEL namespaces."""

    def list_namespaces(self) -> List[Namespace]:
        """List all namespaces."""
        return self._list_model(Namespace)

    def count_namespaces(self) -> int:
        """Count the number of namespaces in the database."""
        return self._count_model(Namespace)

    def count_namespace_entries(self) -> int:
        """Count the number of namespace entries in the database."""
        return self._count_model(NamespaceEntry)

    def drop_namespaces(self):
        """Drop all namespaces."""
        self.session.query(NamespaceEntry).delete()
        self.session.query(Namespace).delete()
        self.session.commit()

    def drop_namespace_by_url(self, url: str) -> None:
        """Drop the namespace at the given URL.

        Won't work if the edge store is in use.

        :param url: The URL of the namespace to drop
        """
        namespace = self.get_namespace_by_url(url)
        self.session.query(NamespaceEntry).filter(NamespaceEntry.namespace == namespace).delete()
        self.session.delete(namespace)
        self.session.commit()

    def get_namespace_by_url(self, url: str) -> Optional[Namespace]:
        """Look up a namespace by url."""
        return self.session.query(Namespace).filter(Namespace.url == url).one_or_none()

    def get_namespace_by_keyword_version(self, keyword: str, version: str) -> Optional[Namespace]:
        """Get a namespace with a given keyword and version."""
        filt = and_(Namespace.keyword == keyword, Namespace.version == version)
        return self.session.query(Namespace).filter(filt).one_or_none()

    def _ensure_namespace_urls(
        self,
        urls: Iterable[str],
        use_tqdm: bool = True,
        is_annotation: bool = False,
    ) -> List[Namespace]:
        ext = "belanno" if is_annotation else "belns"

        rv = []
        url_to_namespace = {}
        url_to_values = {}
        url_to_name_to_id = {}

        tag = "annotations" if is_annotation else "namespaces"

        if use_tqdm:
            urls = tqdm(urls, desc=f"downloading {tag}")
        for url in urls:
            result = self.get_namespace_by_url(url)
            if result:
                rv.append(result)
                continue
            bel_resource = get_bel_resource(url)
            _clean_bel_namespace_values(bel_resource)
            url_to_values[url] = bel_resource["Values"]

            if is_annotation:
                namespace_kwargs = _get_annotation_insert_values(bel_resource)
            else:
                namespace_kwargs = _get_namespace_insert_values(bel_resource)
            result = url_to_namespace[url] = Namespace(url=url, **namespace_kwargs)
            rv.append(result)
            if url.endswith(f"-names.{ext}"):
                mapping_url = url[: -len(f"-names.{ext}")] + f".{ext}.mapping"
                try:
                    res = requests.get(mapping_url)
                    res.raise_for_status()
                except requests.exceptions.HTTPError:
                    logger.warning("No mappings found for %s", url)
                else:
                    mappings = res.json()
                    logger.debug("got %d mappings", len(mappings))
                    url_to_name_to_id[url] = {v: k for k, v in res.json().items()}

        self.session.add_all(url_to_namespace.values())
        self.session.commit()

        url_to_id = {url: namespace.id for url, namespace in url_to_namespace.items()}

        if not url_to_values:
            return rv

        rows = []
        it = url_to_values.items()
        if use_tqdm:
            it = tqdm(it, desc=f"making {tag} entry table")
        if is_annotation:
            for url, values in it:
                for name, identifier in values.items():
                    if not name:
                        continue
                    rows.append((url_to_id[url], name, None, identifier))  # TODO is this a fair assumption?
        else:
            for url, values in it:
                name_to_id = url_to_name_to_id.get(url, {})
                for name, encoding in values.items():
                    if not name:
                        continue
                    rows.append((url_to_id[url], name, encoding, name_to_id.get(name)))

        df = pd.DataFrame(rows, columns=["namespace_id", "name", "encoding", "identifier"])
        logger.info("preparing sql objects for %s", tag)
        df.to_sql(
            NamespaceEntry.__tablename__,
            con=self.engine,
            if_exists="append",
            index=False,
        )
        logger.info("committing %s", tag)
        start_commit_time = time.time()
        self.session.commit()
        logger.info(
            "done committing %s after %.2f seconds",
            tag,
            time.time() - start_commit_time,
        )

        return rv

    def get_or_create_namespace(self, url: str) -> Namespace:
        """Insert the namespace file at the given location to the cache.

        If not cachable, returns the dict of the values of this namespace.

        :raises: pybel.resources.exc.ResourceError
        """
        return self._ensure_namespace_urls([url])[0]

    def get_namespace_by_keyword_pattern(self, keyword: str, pattern: str) -> Optional[Namespace]:
        """Get a namespace with a given keyword and pattern."""
        filt = and_(Namespace.keyword == keyword, Namespace.pattern == pattern)
        return self.session.query(Namespace).filter(filt).one_or_none()

    def ensure_regex_namespace(self, keyword: str, pattern: str) -> Namespace:
        """Get or create a regular expression namespace.

        :param keyword: The keyword of a regular expression namespace
        :param pattern: The pattern for a regular expression namespace
        """
        if pattern is None:
            raise ValueError("cannot have null pattern")

        namespace = self.get_namespace_by_keyword_pattern(keyword, pattern)

        if namespace is None:
            logger.info("creating regex namespace: %s:%s", keyword, pattern)
            namespace = Namespace(
                keyword=keyword,
                pattern=pattern,
            )
            self.session.add(namespace)
            self.session.commit()

        return namespace

    def get_namespace_entry(self, url: str, name: str) -> Optional[NamespaceEntry]:
        """Get a given NamespaceEntry object.

        :param url: The url of the namespace source
        :param name: The value of the namespace from the given url's document
        """
        entry_filter = and_(Namespace.url == url, NamespaceEntry.name == name)
        result = self.session.query(NamespaceEntry).join(Namespace).filter(entry_filter).all()

        if 0 == len(result):
            logger.debug("could not find namespace entry for %s in url=%s", name, url)
            return

        if 1 < len(result):
            logger.warning(
                "result for get_namespace_entry is too long. Returning first of %s",
                [str(r) for r in result],
            )

        return result[0]

    def get_entity_by_identifier(self, url: str, identifier: str) -> Optional[NamespaceEntry]:
        """Get a given entity by its url/identifier combination."""
        entry_filter = and_(Namespace.url == url, NamespaceEntry.identifier == identifier)
        return self.session.query(NamespaceEntry).join(Namespace).filter(entry_filter).one_or_none()

    def get_or_create_regex_namespace_entry(self, *, pattern: str, concept: Entity) -> NamespaceEntry:
        """Get a namespace entry from a regular expression.

        Need to commit after!

        :param pattern: The regular expression pattern for the namespace
        :param concept: The prefix/identifier/name triple
        """
        namespace = self.ensure_regex_namespace(concept.namespace, pattern)

        n_filter = and_(Namespace.pattern == pattern, NamespaceEntry.name == concept.name)

        namespace_entry = self.session.query(NamespaceEntry).join(Namespace).filter(n_filter).one_or_none()

        if namespace_entry is None:
            namespace_entry = NamespaceEntry(
                namespace=namespace,
                name=concept.name,
                identifier=concept.identifier,
            )
            self.session.add(namespace_entry)

        return namespace_entry

    def list_annotations(self) -> List[Namespace]:
        """List all annotations."""
        return self.session.query(Namespace).filter(Namespace.is_annotation).all()

    def count_annotations(self) -> int:
        """Count the number of annotations in the database."""
        return self.session.query(Namespace).filter(Namespace.is_annotation).count()

    def count_annotation_entries(self) -> int:
        """Count the number of annotation entries in the database."""
        return self.session.query(NamespaceEntry).filter(NamespaceEntry.is_annotation).count()

    def get_or_create_annotation(self, url: str) -> Namespace:
        """Insert the namespace file at the given location to the cache.

        :raises: pybel.resources.exc.ResourceError
        """
        return self._ensure_namespace_urls([url], is_annotation=True)[0]

    def get_annotation_entries_by_names(self, url: str, entities: Iterable[Entity]) -> List[NamespaceEntry]:
        """Get annotation entries by URL and names.

        :param url: The url of the annotation source
        :param entities: The names of the annotation entries from the given url's document
        """
        names = [e.identifier if isinstance(e, Entity) else e for e in entities]
        annotation_filter = and_(Namespace.url == url, NamespaceEntry.name.in_(names))
        return self.session.query(NamespaceEntry).join(Namespace).filter(annotation_filter).all()


class NetworkManager(NamespaceManager):
    """Groups functions for inserting and querying networks in the database's network store."""

    def count_networks(self) -> int:
        """Count the networks in the database."""
        return self._count_model(Network)

    def list_networks(self) -> List[Network]:
        """List all networks in the database."""
        return self._list_model(Network)

    def list_recent_networks(self) -> List[Network]:
        """List the most recently created version of each network (by name)."""
        most_recent_times = self.session.query(
            Network.name.label("network_name"),
            func.max(Network.created).label("max_created"),
        )

        most_recent_times = most_recent_times.group_by(Network.name).subquery("most_recent_times")

        and_condition = and_(
            most_recent_times.c.network_name == Network.name,
            most_recent_times.c.max_created == Network.created,
        )

        most_recent_networks = self.session.query(Network).join(most_recent_times, and_condition)

        return most_recent_networks.all()

    def has_name_version(self, name: str, version: str) -> bool:
        """Check if there exists a network with the name/version combination in the database."""
        return self.session.query(exists().where(and_(Network.name == name, Network.version == version))).scalar()

    def drop_networks(self) -> None:
        """Drop all networks."""
        for network in self.session.query(Network).all():
            self.drop_network(network)

    def drop_network_by_id(self, network_id: int) -> None:
        """Drop a network by its database identifier."""
        network = self.session.query(Network).get(network_id)
        self.drop_network(network)

    def drop_network(self, network: Network) -> None:
        """Drop a network, while also cleaning up any edges that are no longer part of any network."""
        # get the IDs of the edges that will be orphaned by deleting this network
        # FIXME: this list could be a problem if it becomes very large; possible optimization is a temporary table in DB
        edge_ids = [result.edge_id for result in self.query_singleton_edges_from_network(network)]

        # delete the network-to-node mappings for this network
        self.session.query(network_node).filter(network_node.c.network_id == network.id).delete(
            synchronize_session=False,
        )

        # delete the edge-to-annotation mappings for the to-be-orphaned edges
        self.session.query(edge_annotation).filter(edge_annotation.c.edge_id.in_(edge_ids)).delete(
            synchronize_session=False,
        )

        # delete the edge-to-network mappings for this network
        self.session.query(network_edge).filter(network_edge.c.network_id == network.id).delete(
            synchronize_session=False,
        )

        # delete the now-orphaned edges
        self.session.query(Edge).filter(Edge.id.in_(edge_ids)).delete(synchronize_session=False)

        # delete the network
        self.session.query(Network).filter(Network.id == network.id).delete(synchronize_session=False)

        # commit it!
        self.session.commit()

    def query_singleton_edges_from_network(self, network: Network) -> sqlalchemy.orm.query.Query:
        """Return a query selecting all edge ids that only belong to the given network."""
        ne1 = aliased(network_edge, name="ne1")
        ne2 = aliased(network_edge, name="ne2")
        singleton_edge_ids_for_network = (
            self.session.query(ne1.c.edge_id)
            .outerjoin(
                ne2,
                and_(
                    ne1.c.edge_id == ne2.c.edge_id,
                    ne1.c.network_id != ne2.c.network_id,
                ),
            )
            .filter(  # noqa: E131
                and_(
                    ne1.c.network_id == network.id,
                    ne2.c.edge_id == None,  # noqa: E711
                ),
            )
        )
        return singleton_edge_ids_for_network

    def get_network_versions(self, name: str) -> Set[str]:
        """Return all of the versions of a network with the given name."""
        return {version for version, in self.session.query(Network.version).filter(Network.name == name).all()}

    def get_network_by_name_version(self, name: str, version: str) -> Optional[Network]:
        """Load the network with the given name and version if it exists."""
        name_version_filter = and_(Network.name == name, Network.version == version)
        network = self.session.query(Network).filter(name_version_filter).one_or_none()
        return network

    def get_graph_by_name_version(self, name: str, version: str) -> Optional[BELGraph]:
        """Load the BEL graph with the given name, or allows for specification of version."""
        network = self.get_network_by_name_version(name, version)

        if network is None:
            return

        return network.as_bel()

    def get_networks_by_name(self, name: str) -> List[Network]:
        """Get all networks with the given name. Useful for getting all versions of a given network."""
        return self.session.query(Network).filter(Network.name.like(name)).all()

    def get_most_recent_network_by_name(self, name: str) -> Optional[Network]:
        """Get the most recently created network with the given name."""
        return self.session.query(Network).filter(Network.name == name).order_by(Network.created.desc()).first()

    def get_graph_by_most_recent(self, name: str) -> Optional[BELGraph]:
        """Get the most recently created network with the given name as a :class:`pybel.BELGraph`."""
        network = self.get_most_recent_network_by_name(name)

        if network is None:
            return

        return network.as_bel()

    def get_network_by_id(self, network_id: int) -> Network:
        """Get a network from the database by its identifier."""
        return self.session.query(Network).get(network_id)

    def get_graph_by_id(self, network_id: int) -> BELGraph:
        """Get a network from the database by its identifier and converts it to a BEL graph."""
        network = self.get_network_by_id(network_id)
        logger.debug("converting network [id=%d] %s to bel graph", network_id, network)
        return network.as_bel()

    def get_networks_by_ids(self, network_ids: Iterable[int]) -> List[Network]:
        """Get a list of networks with the given identifiers.

        Note: order is not necessarily preserved.
        """
        logger.debug("getting networks by identifiers: %s", network_ids)
        return self.session.query(Network).filter(Network.id_in(network_ids)).all()

    def get_graphs_by_ids(self, network_ids: Iterable[int]) -> List[BELGraph]:
        """Get a list of networks with the given identifiers and converts to BEL graphs."""
        rv = [self.get_graph_by_id(network_id) for network_id in network_ids]
        logger.debug("returning graphs for network identifiers: %s", network_ids)
        return rv

    def get_graph_by_ids(self, network_ids: List[int]) -> BELGraph:
        """Get a combine BEL Graph from a list of network identifiers."""
        if len(network_ids) == 1:
            return self.get_graph_by_id(network_ids[0])

        logger.debug("getting graph by identifiers: %s", network_ids)
        graphs = self.get_graphs_by_ids(network_ids)

        logger.debug("getting union of graphs: %s", network_ids)
        rv = union(graphs)

        return rv


class InsertManager(NamespaceManager, LookupManager):
    """Manages inserting data into the edge store."""

    def __init__(self, *args, **kwargs):
        super(InsertManager, self).__init__(*args, **kwargs)

        # A set of dictionaries that contains objects of the type described by the key
        self.object_cache_modification = {}
        self.object_cache_property = {}
        self.object_cache_node = {}
        self.object_cache_edge = {}
        self.object_cache_evidence = {}
        self.curie_to_citation = {}
        self.object_cache_author = {}

    def insert_graph(
        self,
        graph: BELGraph,
        use_tqdm: bool = True,
    ) -> Network:
        """Insert a graph in the database and returns the corresponding Network model.

        :raises: pybel.resources.exc.ResourceError
        """
        if not graph.name:
            raise ValueError("Can not upload a graph without a name")

        if not graph.version:
            raise ValueError("Can not upload a graph without a version")

        logger.debug("inserting %s v%s", graph.name, graph.version)

        t = time.time()

        namespace_urls = graph.namespace_url.values()
        self._ensure_namespace_urls(namespace_urls, use_tqdm=use_tqdm)
        for keyword, pattern in graph.namespace_pattern.items():
            self.ensure_regex_namespace(keyword, pattern)

        annotation_urls = graph.annotation_url.values()
        self._ensure_namespace_urls(annotation_urls, use_tqdm=use_tqdm, is_annotation=True)

        network = Network(**{key: value for key, value in graph.document.items() if key in METADATA_INSERT_KEYS})

        network.store_bel(graph)

        network.nodes, network.edges = self._store_graph_parts(graph, use_tqdm=use_tqdm)

        self.session.add(network)
        self.session.commit()

        logger.info(
            "inserted %s v%s in %.2f seconds",
            graph.name,
            graph.version,
            time.time() - t,
        )

        return network

    def _store_graph_parts(self, graph: BELGraph, use_tqdm: bool = False) -> Tuple[List[Node], List[Edge]]:
        """Store the given graph into the edge store.

        :raises: pybel.resources.exc.ResourceError
        :raises: EdgeAddError
        """
        logger.debug("inserting %s into edge store", graph)
        logger.debug("building node models")
        node_model_build_start = time.time()

        nodes = list(graph)
        if use_tqdm:
            nodes = tqdm(nodes, total=graph.number_of_nodes(), desc="nodes")

        node_model = {}
        for node in nodes:
            node_object = self.get_or_create_node(graph, node)

            if node_object is None:
                logger.warning("can not add node %s", node)
                continue

            node_model[node] = node_object

        node_models = list(node_model.values())
        logger.debug(
            "built %d node models in %.2f seconds",
            len(node_models),
            time.time() - node_model_build_start,
        )

        node_model_commit_start = time.time()
        self.session.add_all(node_models)
        self.session.commit()
        logger.debug(
            "stored %d node models in %.2f seconds",
            len(node_models),
            time.time() - node_model_commit_start,
        )

        logger.debug("building edge models")
        edge_model_build_start = time.time()

        edges = graph.edges(keys=True, data=True)
        if use_tqdm:
            edges = tqdm(edges, total=graph.number_of_edges(), desc="edges")

        edge_models = list(self._get_edge_models(graph, node_model, edges))

        logger.debug(
            "built %d edge models in %.2f seconds",
            len(edge_models),
            time.time() - edge_model_build_start,
        )

        edge_model_commit_start = time.time()
        self.session.add_all(edge_models)
        self.session.commit()
        logger.debug(
            "stored %d edge models in %.2f seconds",
            len(edge_models),
            time.time() - edge_model_commit_start,
        )

        return node_models, edge_models

    def _get_edge_models(
        self,
        graph: BELGraph,
        tuple_model: Mapping[BaseEntity, Node],
        edges,
    ) -> Iterable[Edge]:
        for u, v, key, data in edges:
            source = tuple_model.get(u)
            if source is None or source.md5 not in self.object_cache_node:
                logger.warning("skipping uncached source node: %s", u)
                continue

            target = tuple_model.get(v)
            if target is None or target.md5 not in self.object_cache_node:
                logger.warning("skipping uncached target node: %s", v)
                continue

            relation = data[RELATION]

            if relation in UNQUALIFIED_EDGES:
                try:
                    edge = self._add_unqualified_edge(
                        source=source,
                        target=target,
                        bel=graph.edge_to_bel(u, v, data),
                        key=key,
                        data=data,
                    )
                    if edge is None:
                        continue
                except Exception as e:
                    self.session.rollback()
                    logger.exception("error storing edge in database. edge data: %s", data)
                    raise EdgeAddError(e, u, v, key, data) from e
                else:
                    yield edge

            elif EVIDENCE not in data or CITATION not in data:
                continue

            elif NAMESPACE not in data[CITATION] or IDENTIFIER not in data[CITATION]:
                continue

            else:
                try:
                    bel = graph.edge_to_bel(u, v, data)
                    edge = self._add_qualified_edge(
                        graph=graph,
                        source=source,
                        target=target,
                        key=key,
                        bel=bel,
                        data=data,
                    )
                    if edge is None:
                        continue
                except Exception as e:
                    self.session.rollback()
                    logger.exception("error storing edge in database. edge data: %s", data)
                    raise EdgeAddError(e, u, v, key, data)
                else:
                    yield edge

    @staticmethod
    def _iter_from_annotations_dict(
        graph: BELGraph,
        annotations_dict: AnnotationsDict,
    ) -> Iterable[Tuple[str, Set[Entity]]]:
        """Iterate over the key/value pairs in this edge data dictionary normalized to their source URLs."""
        for key, entities in annotations_dict.items():
            if key in graph.annotation_url:
                url = graph.annotation_url[key]
            elif key in graph.annotation_list:
                continue  # skip those
            elif key in graph.annotation_pattern:
                logger.debug("pattern annotation in database not implemented yet not implemented")  # FIXME
                continue
            else:
                raise ValueError("Graph resources does not contain keyword: {}".format(key))

            yield url, set(entities)

    def _get_annotation_entries_from_data(self, graph: BELGraph, data: EdgeData) -> Optional[List[NamespaceEntry]]:
        """Get the annotation entries from an edge data dictionary."""
        annotations_dict = data.get(ANNOTATIONS)
        if annotations_dict is None:
            return
        rv = []
        for url, entities in self._iter_from_annotations_dict(graph, annotations_dict=annotations_dict):
            for entry in self.get_annotation_entries_by_names(url, entities):
                rv.append(entry)
        return rv

    def _add_qualified_edge(
        self,
        graph: BELGraph,
        source: Node,
        target: Node,
        key: str,
        bel: str,
        data: EdgeData,
    ) -> Optional[Edge]:
        """Add a qualified edge to the network."""
        citation_dict = data[CITATION]
        citation = self.get_or_create_citation(
            namespace=citation_dict[NAMESPACE],
            identifier=citation_dict[IDENTIFIER],
        )
        evidence = self.get_or_create_evidence(citation, data[EVIDENCE])
        annotations = self._get_annotation_entries_from_data(graph, data)

        return self.get_or_create_edge(
            source=source,
            target=target,
            relation=data[RELATION],
            bel=bel,
            md5=key,
            data=data,
            evidence=evidence,
            annotations=annotations,
        )

    def _add_unqualified_edge(self, source: Node, target: Node, key: str, bel: str, data: EdgeData) -> Edge:
        """Add an unqualified edge to the network."""
        return self.get_or_create_edge(
            source=source,
            target=target,
            relation=data[RELATION],
            bel=bel,
            md5=key,
            data=data,
        )

    def get_or_create_evidence(self, citation: Citation, text: str) -> Evidence:
        """Create an entry and object for given evidence if it does not exist."""
        evidence_tuple = citation.db, citation.db_id, text
        if evidence_tuple in self.object_cache_evidence:
            evidence = self.object_cache_evidence[evidence_tuple]
            self.session.add(evidence)
            return evidence

        evidence = self.get_evidence_by_citation_text(citation, text)
        if evidence is not None:
            self.object_cache_evidence[evidence_tuple] = evidence
            return evidence

        self.object_cache_evidence[evidence_tuple] = evidence = Evidence(
            citation=citation,
            text=text,
        )

        self.session.add(evidence)
        return evidence

    def get_or_create_node(self, graph: BELGraph, node: BaseEntity) -> Optional[Node]:
        """Create an entry and object for given node if it does not exist."""
        node_md5 = node.md5
        if node_md5 in self.object_cache_node:
            return self.object_cache_node[node_md5]

        node_model = self.get_node_by_hash(node_md5)
        if node_model is not None:
            self.object_cache_node[node_md5] = node_model
            return node_model

        node_model = Node._start_from_base_entity(node)

        if not isinstance(node, BaseConcept):
            self.session.add(node_model)
            self.object_cache_node[node_md5] = node_model
            return node_model

        if node.namespace in graph.namespace_url:
            url = graph.namespace_url[node.namespace]
            name = node.name
            entry = self.get_namespace_entry(url, name)

            if entry is None:
                logger.debug(
                    "skipping node with entity %s:%s from url=%s",
                    node.namespace,
                    name,
                    url,
                )
                return

            self.session.add(entry)
            node_model.namespace_entry = entry

        elif node.namespace in graph.namespace_pattern:
            entry = self.get_or_create_regex_namespace_entry(
                concept=node.entity,
                pattern=graph.namespace_pattern[node.namespace],
            )
            self.session.add(entry)
            node_model.namespace_entry = entry

        else:
            logger.warning("No reference in BELGraph for namespace: {}".format(node.namespace))
            return

        self.session.add(node_model)
        self.object_cache_node[node_md5] = node_model
        return node_model

    def drop_nodes(self) -> None:
        """Drop all nodes in the database."""
        t = time.time()

        self.session.query(Node).delete()
        self.session.commit()

        logger.info("dropped all nodes in %.2f seconds", time.time() - t)

    def drop_edges(self) -> None:
        """Drop all edges in the database."""
        t = time.time()

        self.session.query(Edge).delete()
        self.session.commit()

        logger.info("dropped all edges in %.2f seconds", time.time() - t)

    def get_or_create_edge(
        self,
        source: Node,
        target: Node,
        relation: str,
        bel: str,
        md5: str,
        data: EdgeData,
        evidence: Optional[Evidence] = None,
        annotations: Optional[List[NamespaceEntry]] = None,
    ) -> Edge:
        """Create an edge if it does not exist, or return it if it does.

        :param source: Source node of the relation
        :param target: Target node of the relation
        :param relation: Type of the relation between source and target node
        :param bel: BEL statement that describes the relation
        :param md5: The MD5 hash of the edge as a string
        :param data: The PyBEL data dictionary
        :param evidence: Evidence object that proves the given relation
        :param annotations: List of all annotations that belong to the edge
        """
        if md5 in self.object_cache_edge:
            edge = self.object_cache_edge[md5]
            self.session.add(edge)
            return edge

        edge = self.get_edge_by_hash(md5)

        if edge is not None:
            self.object_cache_edge[md5] = edge
            return edge

        edge = Edge(
            source=source,
            source_modifier=data.get(SOURCE_MODIFIER),
            target=target,
            target_modifier=data.get(TARGET_MODIFIER),
            relation=relation,
            bel=bel,
            md5=md5,
            data=data,
        )

        if evidence is not None:
            edge.evidence = evidence
        if annotations is not None:
            edge.annotations = annotations

        self.session.add(edge)
        self.object_cache_edge[md5] = edge
        return edge

    def get_or_create_citation(
        self,
        *,
        identifier: str,
        namespace: Optional[str] = None,
    ) -> Citation:
        """Create a citation if it does not exist, or return it if it does.

        :param identifier: Identifier of the given citation (e.g. PubMed id)
        :param namespace: Citation type (defaults to PubMed)
        """
        if namespace is None:
            namespace = CITATION_TYPE_PUBMED

        citation_curie = f"{namespace}:{identifier}"
        if citation_curie in self.curie_to_citation:
            citation = self.curie_to_citation[citation_curie]
            self.session.add(citation)
            return citation

        citation = self.get_citation_by_reference(namespace, identifier)

        if citation is not None:
            self.curie_to_citation[citation_curie] = citation
            return citation

        self.curie_to_citation[citation_curie] = citation = Citation(db=namespace, db_id=identifier)
        self.session.add(citation)
        return citation

    def get_or_create_author(self, name: str) -> Author:
        """Get an author by name, or creates one if it does not exist."""
        author = self.object_cache_author.get(name)

        if author is not None:
            self.session.add(author)
            return author

        author = self.get_author_by_name(name)

        if author is not None:
            self.object_cache_author[name] = author
            return author

        author = self.object_cache_author[name] = Author(name=name)
        self.session.add(author)
        return author


class _Manager(QueryManager, InsertManager, NetworkManager):
    """A wrapper around PyBEL managers that can be directly instantiated with an engine and session."""

    def count_citations(self) -> int:
        """Count the number of citations stored in the database."""
        return self._count_model(Citation)

    def list_citations(self) -> List[Citation]:
        """List the citations in the database."""
        return self._list_model(Citation)


class Manager(_Manager):
    """A manager for the PyBEL database."""

    def __init__(self, connection: Optional[str] = None, engine=None, session=None, **kwargs) -> None:
        """Create a connection to database and a persistent session using SQLAlchemy.

        A custom default can be set as an environment variable with the name :data:`pybel.constants.PYBEL_CONNECTION`,
        using an `RFC-1738 <http://rfc.net/rfc1738.html>`_ string. For example, a MySQL string can be given with the
        following form:

        :code:`mysql+pymysql://<username>:<password>@<host>/<dbname>?charset=utf8[&<options>]`

        A SQLite connection string can be given in the form:

        ``sqlite:///~/Desktop/cache.db``

        Further options and examples can be found on the SQLAlchemy documentation on
        `engine configuration <http://docs.sqlalchemy.org/en/latest/core/engines.html>`_.

        :param connection: An RFC-1738 database connection string. If ``None``, tries to load from the
         environment variable ``PYBEL_CONNECTION`` then from the config file ``~/.config/pybel/config.json`` whose
         value for ``PYBEL_CONNECTION`` defaults to :data:`pybel.constants.DEFAULT_CACHE_CONNECTION`.
        :param engine: Optional engine to use. Must be specified with a session and no connection.
        :param session: Optional session to use. Must be specified with an engine and no connection.
        :param bool echo: Turn on echoing sql
        :param Optional[bool] autoflush: Defaults to True if not specified in kwargs or configuration.
        :param Optional[bool] autocommit: Defaults to False if not specified in kwargs or configuration.
        :param Optional[bool] expire_on_commit: Defaults to False if not specified in kwargs or configuration.
        :param scopefunc: Scoped function to pass to :func:`sqlalchemy.orm.scoped_session`

        From the Flask-SQLAlchemy documentation:

        An extra key ``'scopefunc'`` can be set on the ``options`` dict to
        specify a custom scope function.  If it's not provided, Flask's app
        context stack identity is used. This will ensure that sessions are
        created and removed with the request/response cycle, and should be fine
        in most cases.

        Allowed Usages:

        Instantiation with connection string as positional argument

        >>> my_connection = 'sqlite:///~/Desktop/cache.db'
        >>> manager = Manager(my_connection)

        Instantiation with connection string as positional argument with keyword arguments

        >>> my_connection = 'sqlite:///~/Desktop/cache.db'
        >>> manager = Manager(my_connection, echo=True)

        Instantiation with connection string as keyword argument

        >>> my_connection = 'sqlite:///~/Desktop/cache.db'
        >>> manager = Manager(connection=my_connection)

        Instantiation with connection string as keyword argument with keyword arguments

        >>> my_connection = 'sqlite:///~/Desktop/cache.db'
        >>> manager = Manager(connection=my_connection, echo=True)

        Instantiation with user-supplied engine and session objects as keyword arguments

        >>> my_engine, my_session = ...  # magical creation! See SQLAlchemy documentation
        >>> manager = Manager(engine=my_engine, session=my_session)
        """
        if connection and (engine or session):
            raise ValueError("can not specify connection with engine/session")

        if engine is None and session is None:
            if connection is None:
                connection = get_cache_connection()

            engine, session = build_engine_session(connection=connection, **kwargs)

        elif engine is None or session is None:
            raise ValueError("need both engine and session to be specified")

        elif kwargs:
            raise ValueError("keyword arguments should not be used with engine/session")

        super().__init__(engine=engine, session=session)
        self.create_all()



# -*- coding: utf-8 -*-

"""Tests for node utilities."""

import unittest

from pybel import BELGraph
from pybel.constants import INCREASES
from pybel.dsl import ComplexAbundance as g
from pybel.dsl import CompositeAbundance as c
from pybel.dsl import Protein, Reaction
from pybel.examples.various_example import (
    adp,
    atp,
    glucose,
    glucose_6_phosphate,
    hk1,
    phosphate,
    single_reaction_graph,
)
from pybel.struct.node_utils import flatten_list_abundance, reaction_cartesian_expansion


class TestNodeUtils(unittest.TestCase):
    """Test node utilities."""

    def test_flatten_complex(self):
        """Test flattening a nested complex."""
        p1, p2, p3 = (Protein("N", str(i + 1)) for i in range(3))

        pairs = [
            # Mainly complexes
            (g([p1, p2, p3]), g([p1, p2, p3])),  # no nesting
            (g([p1, p2, p3]), g([g([p1, p2]), p3])),  # one nesting
            (g([p1, p2, p3]), g([g([p1]), p2, p3])),  # one nesting
            (g([p1, p2, p3]), g([g([p1]), g([p2]), p3])),  # one nesting
            # Mainly composites
            (c([p1, p2, p3]), c([p1, p2, p3])),  # no nesting
            (c([p1, p2, p3]), c([c([p1, p2]), p3])),  # one nesting
            (c([p1, p2, p3]), c([c([p1]), p2, p3])),  # one nesting
            (c([p1, p2, p3]), c([c([p1]), c([p2]), p3])),  # one nesting
            # TODO: mixtures of composites and complexes?
        ]

        for expected, source in pairs:
            self.assertEqual(expected, flatten_list_abundance(source))

    def test_flatten_reaction(self):
        """Test flattening a reaction."""
        single_reaction_graph_copy = single_reaction_graph.copy()

        self.assertEqual(single_reaction_graph_copy.number_of_nodes(), 7)
        self.assertEqual(single_reaction_graph_copy.number_of_edges(), 7)

        reaction_cartesian_expansion(single_reaction_graph_copy)

        self.assertEqual(single_reaction_graph_copy.number_of_nodes(), 6)
        self.assertEqual(single_reaction_graph_copy.number_of_edges(), 8)

        pairs = [
            (glucose, INCREASES, glucose_6_phosphate),
            (glucose, INCREASES, adp),
            (hk1, INCREASES, glucose_6_phosphate),
            (hk1, INCREASES, adp),
            (atp, INCREASES, glucose_6_phosphate),
            (atp, INCREASES, adp),
            (phosphate, INCREASES, glucose_6_phosphate),
            (phosphate, INCREASES, adp),
        ]

        for source, target, data in single_reaction_graph_copy.edges(data=True):
            self.assertIn((source, INCREASES, target), pairs)

    def test_flatten_reaction_2(self):
        """Test flattening a qualified reaction."""
        node_increases_reaction_graph = BELGraph()

        glycolisis_step_1 = Reaction(reactants=[glucose, hk1, atp], products=[glucose_6_phosphate, adp, hk1])

        node_increases_reaction_graph.add_increases(glucose_6_phosphate, glycolisis_step_1, citation="X", evidence="X")

        self.assertEqual(node_increases_reaction_graph.number_of_nodes(), 6)
        self.assertEqual(node_increases_reaction_graph.number_of_edges(), 7)

        reaction_cartesian_expansion(node_increases_reaction_graph)

        self.assertEqual(node_increases_reaction_graph.number_of_nodes(), 5)
        # TODO Fix so unqualified duplicate edges are not created (it should be the 8 edges below)
        self.assertEqual(node_increases_reaction_graph.number_of_edges(), 12)

        # pairs = [
        #     (glucose, INCREASES, glucose_6_phosphate),
        #     (glucose, INCREASES, adp),
        #     (hk1, INCREASES, glucose_6_phosphate),
        #     (hk1, INCREASES, adp),
        #     (atp, INCREASES, glucose_6_phosphate),
        #     (atp, INCREASES, adp),
        #     (phosphate, INCREASES, glucose_6_phosphate),
        #     (phosphate, INCREASES, adp),
        # ]
        #
        # for source, target, data in node_increases_reaction_graph.edges(data=True):
        #     self.assertIn((source, INCREASES, target), pairs)

    def test_flatten_reaction_3(self):
        """Test flattening a graph containing 2 reactions connected to each other."""
        two_reactions_graph = BELGraph()

        reaction_1 = Reaction(reactants=[glucose, atp], products=hk1)
        reaction_2 = Reaction(reactants=glucose_6_phosphate, products=adp)

        two_reactions_graph.add_increases(reaction_1, reaction_2, citation="X", evidence="X")

        self.assertEqual(two_reactions_graph.number_of_nodes(), 7)
        self.assertEqual(two_reactions_graph.number_of_edges(), 6)

        reaction_cartesian_expansion(two_reactions_graph)

        # TODO Fix so unqualified duplicate edges are not created (it should be the 6 edges below)
        self.assertEqual(two_reactions_graph.number_of_nodes(), 5)
        self.assertEqual(two_reactions_graph.number_of_edges(), 8)



# -*- coding: utf-8 -*-

"""Tests for parsing utilities."""

import unittest

import networkx as nx

from pybel.utils import subdict_matches
from tests.constants import any_subdict_matches


class TestSubdictMatching(unittest.TestCase):
    """Tests for matching sub-dictionaries."""

    def test_dict_matches_1(self):
        target = {"k1": "v1", "k2": "v2"}
        query = {"k1": "v1", "k2": "v2"}
        self.assertTrue(subdict_matches(target, query))

    def test_dict_matches_2(self):
        target = {"k1": "v1", "k2": "v2", "k3": "v3"}
        query = {"k1": "v1", "k2": "v2"}
        self.assertTrue(subdict_matches(target, query))

    def test_dict_matches_3(self):
        target = {
            "k1": "v1",
        }
        query = {"k1": "v1", "k2": "v2"}
        self.assertFalse(subdict_matches(target, query))

    def test_dict_matches_4(self):
        target = {"k1": "v1", "k2": "v4", "k3": "v3"}
        query = {"k1": "v1", "k2": "v2"}
        self.assertFalse(subdict_matches(target, query))

    def test_dict_matches_5(self):
        target = {"k1": "v1", "k2": "v2"}
        query = {"k1": "v1", "k2": ["v2", "v3"]}
        self.assertTrue(subdict_matches(target, query))

    def test_dict_matches_6(self):
        target = {"k1": "v1", "k2": ["v2", "v3"]}
        query = {"k1": "v1", "k2": "v4"}
        self.assertFalse(subdict_matches(target, query))

    def test_dict_matches_7_partial(self):
        """Tests a partial match"""
        target = {"k1": "v1", "k2": "v2"}
        query = {"k1": "v1", "k2": {"v2": "v3"}}

        self.assertFalse(subdict_matches(target, query))

    def test_dict_matches_7_exact(self):
        """Test a partial match."""
        target = {"k1": "v1", "k2": "v2"}
        query = {"k1": "v1", "k2": {"v2": "v3"}}

        self.assertFalse(subdict_matches(target, query, partial_match=False))

    def test_dict_matches_8_partial(self):
        """Test a partial match."""
        target = {"k1": "v1", "k2": {"v2": "v3", "v4": "v5"}}
        query = {"k1": "v1", "k2": {"v2": "v3"}}

        self.assertTrue(subdict_matches(target, query))

    def test_dict_matches_graph(self):
        """Test matching a graph."""
        g = nx.MultiDiGraph()

        g.add_node(1)
        g.add_node(2)
        g.add_edge(1, 2, relation="yup")
        g.add_edge(1, 2, relation="nope")

        d = {"relation": "yup"}

        self.assertTrue(any_subdict_matches(g[1][2], d))



# -*- coding: utf-8 -*-

"""Tests for PyNPA."""

import unittest

import pandas as pd

from pybel import BELGraph
from pybel.dsl import ComplexAbundance, Gene, Protein, Rna
from pybel.io.pynpa import to_npa_dfs, to_npa_layers
from pybel.struct.getters import get_tf_pairs
from pybel.testing.utils import n

g1 = Gene("hgnc", "1")
r1 = Rna("hgnc", "1")
p1 = Protein("hgnc", "1")
g2 = Gene("hgnc", "2")
r2 = Rna("hgnc", "2")
p2 = Protein("hgnc", "2")
g3 = Gene("hgnc", "3")
p3 = Protein("hgnc", "3")


class TestPyNPA(unittest.TestCase):
    """Tests for PyNPA."""

    def setUp(self) -> None:
        """Set up a small test graph."""
        self.graph = BELGraph()
        self.graph.add_increases(ComplexAbundance([p1, g2]), r2, citation=n(), evidence=n())
        self.graph.add_increases(p2, p3, citation=n(), evidence=n())

    def test_get_tf_pairs(self):
        """Test getting transcription factor-target pairs."""
        tf_pairs = set(get_tf_pairs(self.graph))
        self.assertLess(0, len(tf_pairs), msg="No TF pairs pairs found")
        self.assertEqual(1, len(tf_pairs))
        example_tf, example_target, _ = list(tf_pairs)[0]
        self.assertEqual(p1, example_tf)
        self.assertEqual(r2, example_target)

    def test_export_layers(self):
        """Test that the layers are exported right."""
        ppi_layer, tf_layer = to_npa_layers(self.graph)
        self.assertIsInstance(ppi_layer, dict)
        self.assertIsInstance(tf_layer, dict)
        self.assertLess(0, len(ppi_layer), msg="PPI layer was not populated")
        self.assertLess(0, len(tf_layer), msg="TF layer was not populated")
        self.assertIn((g2, g3), ppi_layer)
        self.assertEqual(+1, ppi_layer[g2, g3])
        self.assertIn((g1, g2), tf_layer)
        self.assertEqual(+1, tf_layer[g1, g2])

    def test_export_df_curie(self):
        """Test exporting dataframes with curie-based nomenclature."""
        ppi_df, tf_df = to_npa_dfs(self.graph)
        self.assertIsInstance(ppi_df, pd.DataFrame)
        self.assertIsInstance(tf_df, pd.DataFrame)
        self.assertEqual(1, len(ppi_df.index))
        self.assertEqual(1, len(tf_df.index))
        self.assertEqual(("hgnc:2", "hgnc:3", 1), tuple(list(ppi_df.values)[0]))
        self.assertEqual(("hgnc:1", "hgnc:2", 1), tuple(list(tf_df.values)[0]))

    def test_export_df_name(self):
        """Test exporting dataframes with name-based nomenclature."""
        ppi_df, tf_df = to_npa_dfs(
            self.graph,
            nomenclature_method_first_layer="name",
            nomenclature_method_second_layer="name",
        )
        self.assertEqual(("2", "3", 1), tuple(list(ppi_df.values)[0]))
        self.assertEqual(("1", "2", 1), tuple(list(tf_df.values)[0]))

    def test_export_df_inode(self):
        """Test exporting dataframes with inode-based nomenclature in the second layer."""
        ppi_df, tf_df = to_npa_dfs(
            self.graph,
            nomenclature_method_first_layer="name",
            nomenclature_method_second_layer="inodes",
        )
        self.assertEqual(("2", "3", 1), tuple(list(ppi_df.values)[0]))
        self.assertEqual(("*1", "*2", 1), tuple(list(tf_df.values)[0]))



# -*- coding: utf-8 -*-

"""The Umbrella Node-Link JSON format is similar to node-link but uses full BEL terms as nodes.

Given a BEL statement describing that ``X`` phosphorylates ``Y`` like ``act(p(X)) -> p(Y, pmod(Ph))``,
PyBEL usually stores the ``act()`` information about ``X`` as part of the relationship. In Umbrella mode,
this stays as part of the node.

Note that this generates additional nodes in the network for each of the "modified" versions of
the node. For example, ``act(p(X))`` will be represented as individual node instead of
``p(X)``, as in the standard node-link JSON exporter.

A user might want to use this exporter in the following scenarios:

- Represent transitivity in activities like in ``p(X, pmod(Ph)) -> act(p(X)) -> p(Y, pmod(Ph)) -> act(p(Y))``
  with four nodes that are more ammenable to simulatons (e.g., boolean networks, petri nets).
- Visualizing networks that in similar way to the legacy BEL
  `Cytoscape plugin <https://apps.cytoscape.org/apps/belnavigator>`_ from the BEL Framework (warning: now defunct)
  using tools like Cytoscape.
"""

import gzip
import json
from itertools import chain, count
from typing import Any, Mapping, TextIO, Union

from networkx.utils import open_file

from ..canonicalize import _decanonicalize_edge_node, edge_to_tuple
from ..constants import GRAPH_ANNOTATION_LIST, SOURCE_MODIFIER, TARGET_MODIFIER
from ..struct import BELGraph

__all__ = [
    "to_umbrella_nodelink",
    "to_umbrella_nodelink_file",
    "to_umbrella_nodelink_gz",
]


def to_umbrella_nodelink(graph: BELGraph) -> Mapping[str, Any]:
    """Convert this graph to an umbrella node-link JSON object.

    :param graph: A BEL graph
    """
    nodes = set()

    for u, v, data in graph.edges(data=True):
        u_key, _, v_key = edge_to_tuple(u, v, data)
        nodes.add(u_key)
        nodes.add(v_key)

    nodes = sorted(list(nodes))
    mapping = dict(zip(nodes, count()))

    graph_json_dict = {
        "directed": True,
        "multigraph": True,
        "graph": graph.graph.copy(),
        "nodes": nodes,
        "links": [
            dict(
                chain(
                    data.copy().items(),
                    [
                        (
                            "source",
                            mapping[_decanonicalize_edge_node(u, data, node_position=SOURCE_MODIFIER)],
                        ),
                        (
                            "target",
                            mapping[_decanonicalize_edge_node(v, data, node_position=TARGET_MODIFIER)],
                        ),
                        ("key", key),
                    ],
                ),
            )
            for u, v, key, data in graph.edges(keys=True, data=True)
        ],
    }

    # Convert annotation list definitions (which are sets) to canonicalized/sorted lists
    graph_json_dict["graph"][GRAPH_ANNOTATION_LIST] = {
        keyword: list(sorted(values))
        for keyword, values in graph_json_dict["graph"].get(GRAPH_ANNOTATION_LIST, {}).items()
    }

    return graph_json_dict


@open_file(1, mode="w")
def to_umbrella_nodelink_file(graph: BELGraph, path: Union[str, TextIO], **kwargs) -> None:
    """Write this graph as an umbrella node-link JSON to a file.

    :param graph: A BEL graph
    :param path: A path or file-like
    """
    graph_json_dict = to_umbrella_nodelink(graph)
    json.dump(graph_json_dict, path, ensure_ascii=False, **kwargs)


def to_umbrella_nodelink_gz(graph, path: str, **kwargs) -> None:
    """Write this graph as an umbrella node-link JSON to a gzipped file."""
    with gzip.open(path, "wt") as file:
        to_umbrella_nodelink_file(graph, file, **kwargs)



# -*- coding: utf-8 -*-

"""Functions for inducing graphs by nodes and by edge filters."""

from typing import Iterable, Optional

import networkx as nx

from ..utils import expand_by_edge_filter
from ...filters.edge_predicates import is_causal_relation
from ...filters.node_filters import filter_nodes
from ...filters.typing import EdgePredicates, NodePredicates
from ...graph import BELGraph
from ...operations import subgraph
from ...pipeline import transformation
from ....dsl import BaseEntity

__all__ = [
    "get_subgraph_by_edge_filter",
    "get_subgraph_by_induction",
    "get_subgraph_by_node_filter",
    "get_largest_component",
    "get_causal_subgraph",
]


@transformation
def get_subgraph_by_edge_filter(graph: BELGraph, edge_predicates: Optional[EdgePredicates] = None) -> BELGraph:
    """Induce a sub-graph on all edges that pass the given filters.

    :param graph: A BEL graph
    :param edge_predicates: An edge predicate or list of edge predicates
    :return: A BEL sub-graph induced over the edges passing the given filters
    """
    rv = graph.child()
    expand_by_edge_filter(graph, rv, edge_predicates=edge_predicates)
    return rv


@transformation
def get_subgraph_by_induction(graph: BELGraph, nodes: Iterable[BaseEntity]) -> Optional[BELGraph]:
    """Induce a sub-graph over the given nodes or return None if none of the nodes are in the given graph.

    :param graph: A BEL graph
    :param nodes: A list of BEL nodes in the graph
    """
    nodes = tuple(nodes)

    if all(node not in graph for node in nodes):
        return

    return subgraph(graph, nodes)


@transformation
def get_subgraph_by_node_filter(graph: BELGraph, node_predicates: NodePredicates) -> BELGraph:
    """Induce a sub-graph on the nodes that pass the given predicate(s).

    :param graph: A BEL graph
    :param node_predicates: A node predicate or list of node predicates
    """
    return get_subgraph_by_induction(graph, filter_nodes(graph, node_predicates))


@transformation
def get_largest_component(graph: BELGraph) -> BELGraph:
    """Get the giant component of a graph.

    :param graph: A BEL graph
    """
    biggest_component_nodes = max(nx.weakly_connected_components(graph), key=len)
    return subgraph(graph, biggest_component_nodes)


@transformation
def get_causal_subgraph(graph: BELGraph) -> BELGraph:
    """Build a new sub-graph induced over the causal edges.

    :param graph: A BEL graph
    """
    return get_subgraph_by_edge_filter(graph, is_causal_relation)



# -*- coding: utf-8 -*-

"""Functions for building node predicates."""

from typing import Any, Callable, Iterable, List, Union

from .node_predicates import concatenate_node_predicates, invert_node_predicate
from .typing import NodePredicate
from ..graph import BELGraph
from ...constants import CONCEPT, NAME
from ...dsl import BaseConcept, BaseEntity
from ...typing import Strings

__all__ = [
    "function_inclusion_filter_builder",
    "function_exclusion_filter_builder",
    "data_missing_key_builder",
    "build_node_data_search",
    "build_node_graph_data_search",
    "build_node_key_search",
    "build_node_name_search",
    "namespace_inclusion_builder",
]


def function_inclusion_filter_builder(func: Strings) -> NodePredicate:
    """Build a filter that only passes on nodes of the given function(s).

    :param func: A BEL Function or list/set/tuple of BEL functions
    """
    if isinstance(func, str):
        return _single_function_inclusion_filter_builder(func)

    elif isinstance(func, Iterable):
        return _collection_function_inclusion_builder(func)

    raise TypeError("Invalid type for argument: {}".format(func))


def _single_function_inclusion_filter_builder(func: str) -> NodePredicate:  # noqa: D202
    """Build a function inclusion filter for a single function."""

    def function_inclusion_filter(_: BELGraph, node: BaseEntity) -> bool:
        """Pass only for a node that has the enclosed function."""
        return node.function == func

    return function_inclusion_filter


def _collection_function_inclusion_builder(funcs: Iterable[str]) -> NodePredicate:
    """Build a function inclusion filter for a collection of functions."""
    funcs = set(funcs)

    if not funcs:
        raise ValueError("can not build function inclusion filter with empty list of functions")

    def functions_inclusion_filter(_: BELGraph, node: BaseEntity) -> bool:
        """Pass only for a node that is one of the enclosed functions."""
        return node.function in funcs

    return functions_inclusion_filter


def function_exclusion_filter_builder(func: Strings) -> NodePredicate:
    """Build a filter that fails on nodes of the given function(s).

    :param func: A BEL Function or list/set/tuple of BEL functions
    """
    return invert_node_predicate(function_inclusion_filter_builder(func))


def data_missing_key_builder(key: str) -> NodePredicate:  # noqa: D202
    """Build a filter that passes only on nodes that don't have the given key in their data dictionary.

    :param str key: A key for the node's data dictionary
    """

    def data_does_not_contain_key(graph: BELGraph, node: BaseEntity) -> bool:
        """Pass only for a node that doesn't contain the enclosed key in its data dictionary."""
        return key not in graph.nodes[node]

    return data_does_not_contain_key


def build_node_data_search(  # noqa: D202
    key: Union[str, List[str]],
    data_predicate: Callable[[Any], bool],
) -> NodePredicate:
    """Build a filter for nodes whose associated data with the given key passes the given predicate.

    :param key: The node data dictionary key to check
    :param data_predicate: The filter to apply to the node data dictionary
    """
    if isinstance(key, list):
        _getter = _get_multi_value
    else:
        _getter = _get_single_value

    def node_data_filter(_: BELGraph, node: BaseEntity) -> bool:
        """Pass if the given node has a given data annotated and passes the contained filter."""
        value = _getter(node, key)
        return value is not None and data_predicate(value)

    return node_data_filter


def build_node_graph_data_search(  # noqa: D202
    key: Union[str, List[str]],
    data_predicate: Callable[[Any], bool],
) -> NodePredicate:
    """Build a function for testing data associated with the node in the graph.

    :param key: The node data dictionary key to check
    :param data_predicate: The filter to apply to the node data dictionary
    """
    if isinstance(key, tuple):
        _getter = _get_multi_value
    else:
        _getter = _get_single_value

    def node_data_filter(graph: BELGraph, node: BaseEntity) -> bool:
        """Pass if the given node has a given data annotated and passes the contained filter."""
        value = _getter(graph.nodes[node], key)
        return value is not None and data_predicate(value)

    return node_data_filter


def _get_multi_value(d, keys):
    value = d.get(keys[0])
    for key in keys[1:]:
        if value is None:
            return
        value = value.get(key)
    return value


def _get_single_value(d, key):
    return d.get(key)


def build_node_key_search(
    query: Strings,
    key: Union[str, List[str]],
) -> NodePredicate:
    """Build a node filter for nodes whose values for the given key are superstrings of the query string(s).

    :param query: The query string or strings to check if they're in the node name
    :param key: The key for the node data dictionary. Should refer only to entries that have str values
    """
    if isinstance(query, str):
        return build_node_data_search(key, lambda s: query.lower() in s.lower())

    if isinstance(query, Iterable):
        return build_node_data_search(key, lambda s: any(q.lower() in s.lower() for q in query))

    raise TypeError("query is wrong type: %s", query)


def build_node_name_search(query: Strings) -> NodePredicate:
    """Search nodes' names.

    Is a thin wrapper around :func:`build_node_key_search` with :data:`pybel.constants.NAME`

    :param query: The query string or strings to check if they're in the node name
    """
    return build_node_key_search(query=query, key=[CONCEPT, NAME])


def namespace_inclusion_builder(namespace: Strings) -> NodePredicate:
    """Build a predicate for namespace inclusion."""
    if isinstance(namespace, str):

        def namespace_filter(_: BELGraph, node: BaseEntity) -> bool:
            """Pass only for a node that has the enclosed namespace."""
            return isinstance(node, BaseConcept) and node.namespace == namespace

    elif isinstance(namespace, Iterable):
        namespaces = set(namespace)

        def namespace_filter(_: BELGraph, node: BaseEntity) -> bool:
            """Pass only for a node that has a namespace in the enclosed set."""
            return isinstance(node, BaseConcept) and node.namespace in namespaces

    else:
        raise TypeError("Invalid type for argument: {}".format(namespace))

    return namespace_filter



# -*- coding: utf-8 -*-

"""Tests for interchange with JGIF."""

import json
import logging
import sys
import unittest

from pybel import from_cbn_jgif, to_jgif
from pybel.constants import (
    ACTIVITY,
    ANNOTATIONS,
    CITATION,
    CITATION_TYPE_OTHER,
    CITATION_TYPE_PUBMED,
    DECREASES,
    DIRECTLY_INCREASES,
    EFFECT,
    EVIDENCE,
    IDENTIFIER,
    MODIFIER,
    NAMESPACE,
    RELATION,
    TARGET_MODIFIER,
)
from pybel.dsl import (
    Abundance,
    BiologicalProcess,
    ComplexAbundance,
    NamedComplexAbundance,
    Pathology,
    Protein,
    ProteinModification,
)
from pybel.language import activity_mapping
from pybel.testing.constants import test_jgif_path
from tests.constants import TestGraphMixin

logging.getLogger("pybel.parser").setLevel(20)

calcium = Abundance("SCHEM", "Calcium")
calcineurin_complex = NamedComplexAbundance("SCOMP", "Calcineurin Complex")
foxo3 = Protein("HGNC", "FOXO3")
tcell_proliferation = BiologicalProcess("GO", "CD8-positive, alpha-beta T cell proliferation")
il15 = Protein("HGNC", "IL15")
il2rg = Protein("MGI", "Il2rg")
jgif_expected_nodes = {
    calcium,
    calcineurin_complex,
    foxo3,
    tcell_proliferation,
    il15,
    il2rg,
    Protein("HGNC", "CXCR6"),
    Protein("HGNC", "IL15RA"),
    BiologicalProcess("GO", "lymphocyte chemotaxis"),
    Protein("HGNC", "IL2RG"),
    Protein("HGNC", "ZAP70"),
    NamedComplexAbundance("SCOMP", "T Cell Receptor Complex"),
    BiologicalProcess("GO", "T cell activation"),
    Protein("HGNC", "CCL3"),
    Protein("HGNC", "PLCG1"),
    Protein("HGNC", "FASLG"),
    Protein("HGNC", "IDO1"),
    Protein("HGNC", "IL2"),
    Protein("HGNC", "CD8A"),
    Protein("HGNC", "CD8B"),
    Protein("HGNC", "PLCG1"),
    Protein("HGNC", "BCL2"),
    Protein("HGNC", "CCR3"),
    Protein("HGNC", "IL2RB"),
    Protein("HGNC", "CD28"),
    Pathology("SDIS", "Cytotoxic T-cell activation"),
    Protein("HGNC", "FYN"),
    Protein("HGNC", "CXCL16"),
    Protein("HGNC", "CCR5"),
    Protein("HGNC", "LCK"),
    Protein("SFAM", "Chemokine Receptor Family"),
    Protein("HGNC", "CXCL9"),
    Pathology("SDIS", "T-cell migration"),
    Protein("HGNC", "CXCR3"),
    Abundance("CHEBI", "acrolein"),
    Protein("HGNC", "IDO2"),
    Pathology("MESHD", "Pulmonary Disease, Chronic Obstructive"),
    Protein("HGNC", "IFNG"),
    Protein("HGNC", "TNFRSF4"),
    Protein("HGNC", "CTLA4"),
    Protein("HGNC", "GZMA"),
    Protein("HGNC", "PRF1"),
    Protein("HGNC", "TNF"),
    Protein("SFAM", "Chemokine Receptor Family"),
    ComplexAbundance([Protein("HGNC", "CD8A"), Protein("HGNC", "CD8B")]),
    ComplexAbundance([Protein("HGNC", "CD8A"), Protein("HGNC", "CD8B")]),
    Protein("HGNC", "PLCG1", variants=ProteinModification("Ph", "Tyr")),
    Protein("EGID", "21577"),
}

jgif_expected_edges = [
    (
        calcium,
        calcineurin_complex,
        {
            RELATION: DIRECTLY_INCREASES,
            EVIDENCE: "NMDA-mediated influx of calcium led to activated of the calcium-dependent phosphatase calcineurin and the subsequent dephosphorylation and activation of the protein-tyrosine phosphatase STEP",
            CITATION: {NAMESPACE: CITATION_TYPE_PUBMED, IDENTIFIER: "12483215"},
            TARGET_MODIFIER: {MODIFIER: ACTIVITY, EFFECT: activity_mapping["phos"]},
            ANNOTATIONS: {"Species": {"10116": True}, "Cell": {"neuron": True}},
        },
    ),
    (
        foxo3,
        tcell_proliferation,
        {
            RELATION: DECREASES,
            EVIDENCE: '"These data suggested that FOXO3 downregulates the accumulation of CD8 T cells in tissue specific fashion during an acute LCMV [lymphocytic choriomeningitis virus] infection." (p. 3)',
            CITATION: {NAMESPACE: CITATION_TYPE_OTHER, IDENTIFIER: "22359505"},
            ANNOTATIONS: {
                "Species": {"10090": True},
                "Disease": {"Viral infection": True},
            },
        },
    ),
    (
        il15,
        il2rg,
        {
            RELATION: DIRECTLY_INCREASES,
            EVIDENCE: "IL-15 utilizes ... the common cytokine receptor γ-chain (CD132) for signal transduction in lymphocytes",
            CITATION: {NAMESPACE: CITATION_TYPE_OTHER, IDENTIFIER: "20335267"},
            TARGET_MODIFIER: {MODIFIER: ACTIVITY, EFFECT: activity_mapping["cat"]},
            ANNOTATIONS: {
                "Tissue": {"lung": True},
                "Species": {"9606": True},
            },
        },
    ),
]


class TestJgif(TestGraphMixin):
    """Tests data interchange of JGIF."""

    @unittest.skipIf(sys.platform.startswith("win"), "does not work on Windows")
    def test_jgif_interchange(self):
        """Tests data from CBN"""
        with open(test_jgif_path) as f:
            graph_jgif_dict = json.load(f)

        graph = from_cbn_jgif(graph_jgif_dict)

        self.assertEqual(jgif_expected_nodes, set(graph))

        for u, v, d in jgif_expected_edges:
            self.assert_has_edge(graph, u, v, permissive=False, **d)

        # TODO test more thoroughly?
        export_jgif = to_jgif(graph)
        self.assertIsInstance(export_jgif, dict)



# -*- coding: utf-8 -*-

"""This module contains functions for filtering node and edge iterables.

It relies heavily on the concepts of `functional programming <https://en.wikipedia.org/wiki/Functional_programming>`_
and the concept of `predicates <https://en.wikipedia.org/wiki/Predicate_(mathematical_logic)>`_.
"""

from . import (
    edge_filters,
    edge_predicate_builders,
    edge_predicates,
    node_filters,
    node_predicate_builders,
    node_predicates,
    typing,
    utils,
)
from .edge_filters import *
from .edge_predicate_builders import *
from .edge_predicates import *
from .node_filters import *
from .node_predicate_builders import *
from .node_predicates import *
from .typing import *
from .utils import *

__all__ = [k for k in locals() if not k.startswith("_")]



# -*- coding: utf-8 -*-

"""Convert the Selventa graph for Hipathia."""

import os
from urllib.request import urlretrieve

import click
from pyobo.cli_utils import verbose_option

import pybel
import pybel.grounding

HERE = os.path.dirname(__file__)
URL = 'https://github.com/cthoyt/selventa-knowledge/raw/master/selventa_knowledge/small_corpus.bel.nodelink.json.gz'
PATH = os.path.join(HERE, 'small_corpus.bel.nodelink.json.gz')
GROUNDED_PATH = os.path.join(HERE, 'small_corpus-grounded.bel.nodelink.json.gz')


@click.command()
@verbose_option
def main():
    """Convert the Selventa graph to Hipathia."""
    if not os.path.exists(PATH):
        urlretrieve(URL, PATH)
    if not os.path.exists(GROUNDED_PATH):
        graph = pybel.load(PATH)
        graph = pybel.grounding.ground(graph)
        pybel.dump(graph, GROUNDED_PATH)
    else:
        graph = pybel.load(GROUNDED_PATH)
    pybel.to_hipathia(graph, HERE)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

"""PyBEL implements an internal domain-specific language (DSL).

This enables you to write BEL using Python scripts. Even better,
you can programatically generate BEL using Python. See the
Bio2BEL `paper <https://doi.org/10.1101/631812>`_ and `repository <https://github.com/bio2bel/bio2bel>`_
for many examples.

Internally, the BEL parser converts BEL script into the BEL DSL
then adds it to a BEL graph object. When you iterate through
the :class:`pybel.BELGraph`, the nodes are instances of subclasses
of :class:`pybel.dsl.BaseEntity`.
"""

from .constants import FUNC_TO_DSL, FUNC_TO_FUSION_DSL, FUNC_TO_LIST_DSL
from .edges import (
    activity,
    cell_surface_expression,
    degradation,
    location,
    secretion,
    translocation,
)
from .exc import (
    InferCentralDogmaException,
    ListAbundanceEmptyException,
    PyBELDSLException,
    ReactionEmptyException,
)
from .namespaces import chebi, hgnc, mirbase
from .node_classes import (
    Abundance,
    BaseAbundance,
    BaseConcept,
    BaseEntity,
    BiologicalProcess,
    CentralDogma,
    ComplexAbundance,
    CompositeAbundance,
    Entity,
    EntityVariant,
    EnumeratedFusionRange,
    Fragment,
    FusionBase,
    FusionRangeBase,
    Gene,
    GeneFusion,
    GeneModification,
    Hgvs,
    HgvsReference,
    HgvsUnspecified,
    ListAbundance,
    MicroRna,
    MissingFusionRange,
    NamedComplexAbundance,
    Pathology,
    Population,
    Protein,
    ProteinFusion,
    ProteinModification,
    ProteinSubstitution,
    Reaction,
    Rna,
    RnaFusion,
    Transcribable,
    Variant,
)

entity = Entity

abundance = Abundance

bioprocess = BiologicalProcess
pathology = Pathology

pmod = ProteinModification
gmod = GeneModification
hgvs = Hgvs
hgvs_unspecified = HgvsUnspecified
hgvs_reference = HgvsReference
protein_substitution = ProteinSubstitution
fragment = Fragment

gene = Gene
rna = Rna
mirna = MicroRna
protein = Protein

reaction = Reaction

complex_abundance = ComplexAbundance
named_complex_abundance = NamedComplexAbundance
composite_abundance = CompositeAbundance

missing_fusion_range = MissingFusionRange
fusion_range = EnumeratedFusionRange

protein_fusion = ProteinFusion
rna_fusion = RnaFusion
gene_fusion = GeneFusion



# -*- coding: utf-8 -*-

import logging
import re
import unittest
from random import randint

from pybel.constants import (
    ANNOTATIONS,
    CITATION,
    CITATION_TYPE_PUBMED,
    EVIDENCE,
    IDENTIFIER,
    NAMESPACE,
)
from pybel.exceptions import (
    CitationTooLongException,
    CitationTooShortException,
    IllegalAnnotationValueWarning,
    InvalidCitationType,
    InvalidPubMedIdentifierWarning,
    MissingAnnotationKeyWarning,
    MissingAnnotationRegexWarning,
    UndefinedAnnotationWarning,
)
from pybel.language import Entity
from pybel.parser import ControlParser
from pybel.parser.parse_control import set_citation_stub
from pybel.testing.utils import n
from tests.constants import SET_CITATION_TEST, test_citation_dict

logging.getLogger("requests").setLevel(logging.WARNING)


class TestParseControl(unittest.TestCase):
    def setUp(self):
        self.annotation_to_term = {
            "Custom1": {"Custom1_A", "Custom1_B"},
            "Custom2": {"Custom2_A", "Custom2_B"},
        }

        self.annotation_to_pattern = {"CustomRegex": re.compile("[0-9]+")}

        self.parser = ControlParser(
            annotation_to_term=self.annotation_to_term,
            annotation_to_pattern=self.annotation_to_pattern,
        )


class TestParseControlUnsetStatementErrors(TestParseControl):
    def test_unset_missing_evidence(self):
        with self.assertRaises(MissingAnnotationKeyWarning):
            self.parser.parseString("UNSET Evidence")

    def test_unset_missing_citation(self):
        with self.assertRaises(MissingAnnotationKeyWarning):
            self.parser.parseString("UNSET Citation")

    def test_unset_missing_evidence_with_citation(self):
        """Tests that an evidence can't be unset without a citation"""
        s = [SET_CITATION_TEST, "UNSET Evidence"]
        with self.assertRaises(MissingAnnotationKeyWarning):
            self.parser.parse_lines(s)

    def test_unset_missing_statement_group(self):
        with self.assertRaises(MissingAnnotationKeyWarning):
            self.parser.parseString("UNSET STATEMENT_GROUP")

    def test_unset_missing_command(self):
        s = [SET_CITATION_TEST, "UNSET Custom1"]
        with self.assertRaises(MissingAnnotationKeyWarning):
            self.parser.parse_lines(s)

    def test_unset_invalid_command(self):
        s = [SET_CITATION_TEST, "UNSET MISSING"]
        with self.assertRaises(UndefinedAnnotationWarning):
            self.parser.parse_lines(s)

    def test_unset_list_compact(self):
        """Tests unsetting an annotation list, without spaces in it"""
        s = [
            SET_CITATION_TEST,
            'SET Custom1 = "Custom1_A"',
            'SET Custom2 = "Custom2_A"',
        ]
        self.parser.parse_lines(s)
        self.assertIn("Custom1", self.parser.annotations)
        self.assertIn("Custom2", self.parser.annotations)
        self.parser.parseString("UNSET {Custom1,Custom2}")
        self.assertFalse(self.parser.annotations)

    def test_unset_list_spaced(self):
        """Tests unsetting an annotation list, with spaces in it"""
        s = [
            SET_CITATION_TEST,
            'SET Custom1 = "Custom1_A"',
            'SET Custom2 = "Custom2_A"',
        ]
        self.parser.parse_lines(s)
        self.assertIn("Custom1", self.parser.annotations)
        self.assertIn("Custom2", self.parser.annotations)
        self.parser.parseString("UNSET {Custom1, Custom2}")
        self.assertFalse(self.parser.annotations)


class TestSetCitation(unittest.TestCase):
    def test_parser_double(self):
        set_citation_stub.parseString('Citation = {"PubMed","12928037"}')

    def test_parser_double_spaced(self):
        set_citation_stub.parseString('Citation = {"PubMed", "12928037"}')

    def test_parser_triple(self):
        set_citation_stub.parseString('Citation = {"PubMed Central","Trends in molecular medicine","12928037"}')

    def test_parser_triple_spaced(self):
        set_citation_stub.parseString('Citation = {"PubMed Central", "Trends in molecular medicine", "12928037"}')


class TestParseControlSetStatementErrors(TestParseControl):
    def test_invalid_citation_type(self):
        with self.assertRaises(InvalidCitationType):
            self.parser.parseString('SET Citation = {"PubMedCentral","Trends in molecular medicine","12928037"}')

    def test_invalid_pmid(self):
        with self.assertRaises(InvalidPubMedIdentifierWarning):
            self.parser.parseString('SET Citation = {"PubMed","Trends in molecular medicine","NOT VALID NUMBER"}')

    def test_invalid_pmid_short(self):
        with self.assertRaises(InvalidPubMedIdentifierWarning):
            self.parser.parseString('SET Citation = {"PubMed","NOT VALID NUMBER"}')

    def test_set_missing_statement(self):
        statements = [SET_CITATION_TEST, 'SET MissingKey = "lol"']
        with self.assertRaises(UndefinedAnnotationWarning):
            self.parser.parse_lines(statements)

    def test_custom_annotation_list_withInvalid(self):
        statements = [
            SET_CITATION_TEST,
            'SET Custom1 = {"Custom1_A","Custom1_B","Evil invalid!!!"}',
        ]

        with self.assertRaises(IllegalAnnotationValueWarning):
            self.parser.parse_lines(statements)

    def test_custom_value_failure(self):
        """Tests what happens for a valid annotation key, but an invalid value"""
        s = [SET_CITATION_TEST, 'SET Custom1 = "Custom1_C"']
        with self.assertRaises(IllegalAnnotationValueWarning):
            self.parser.parse_lines(s)

    def test_regex_failure(self):
        s = [SET_CITATION_TEST, 'SET CustomRegex = "abce13"']
        with self.assertRaises(MissingAnnotationRegexWarning):
            self.parser.parse_lines(s)


class TestParseControl2(TestParseControl):
    def test_set_statement_group(self):
        """Tests a statement group gets set properly"""
        s1 = 'SET STATEMENT_GROUP = "my group"'

        self.assertIsNone(self.parser.statement_group)

        self.parser.parseString(s1)
        self.assertEqual("my group", self.parser.statement_group, msg="problem with integration")

        s2 = "UNSET STATEMENT_GROUP"
        self.parser.parseString(s2)
        self.assertIsNone(self.parser.statement_group, msg="problem with unset")

    def test_citation_short(self):
        self.parser.parseString(SET_CITATION_TEST)
        self.assertEqual(test_citation_dict[IDENTIFIER], self.parser.citation_db_id)
        self.assertEqual(test_citation_dict[NAMESPACE], self.parser.citation_db)

        expected_annotations = {
            EVIDENCE: None,
            ANNOTATIONS: {},
            CITATION: test_citation_dict,
        }
        self.assertEqual(expected_annotations, self.parser.get_annotations())

        self.parser.parseString("UNSET Citation")
        self.assertFalse(self.parser.citation_is_set)

    def test_citation_invalid_date(self):
        s = 'SET Citation = {"PubMed","Trends in molecular medicine","12928037","01-12-1999","de Nigris"}'

        self.parser.parseString(s)
        self.assertEqual(CITATION_TYPE_PUBMED, self.parser.citation_db)
        self.assertEqual("12928037", self.parser.citation_db_id)

        expected_dict = {
            EVIDENCE: None,
            ANNOTATIONS: {},
            CITATION: {
                NAMESPACE: CITATION_TYPE_PUBMED,
                IDENTIFIER: "12928037",
            },
        }

        self.assertEqual(expected_dict, self.parser.get_annotations())

    def test_citation_with_empty_comment(self):
        s = 'SET Citation = {"PubMed","Test Name","12928037","1999-01-01","de Nigris|Lerman A|Ignarro LJ",""}'
        self.parser.parseString(s)

        self.assertEqual(CITATION_TYPE_PUBMED, self.parser.citation_db)
        self.assertEqual("12928037", self.parser.citation_db_id)

        expected_dict = {
            EVIDENCE: None,
            ANNOTATIONS: {},
            CITATION: {
                NAMESPACE: CITATION_TYPE_PUBMED,
                IDENTIFIER: "12928037",
            },
        }

        self.assertEqual(expected_dict, self.parser.get_annotations())

    def test_double(self):
        s = 'SET Citation = {"PubMed","12928037"}'
        self.parser.parseString(s)
        self.assertEqual(CITATION_TYPE_PUBMED, self.parser.citation_db)
        self.assertEqual("12928037", self.parser.citation_db_id)

    def test_double_with_space(self):
        """Same as test_double, but has a space between the comma and next entry"""
        s = 'SET Citation = {"PubMed", "12928037"}'
        self.parser.parseString(s)
        self.assertEqual(CITATION_TYPE_PUBMED, self.parser.citation_db)
        self.assertEqual("12928037", self.parser.citation_db_id)

    def test_citation_too_short(self):
        s = 'SET Citation = {"PubMed"}'
        with self.assertRaises(CitationTooShortException):
            self.parser.parseString(s)

    def test_citation_too_long(self):
        s = 'SET Citation = {"PubMed","Name","1234","1999-01-01","Nope|Noper","Nope", "nope nope"}'
        with self.assertRaises(CitationTooLongException):
            self.parser.parseString(s)

    def test_evidence(self):
        self.parser.parseString(SET_CITATION_TEST)
        s = 'SET Evidence = "For instance, during 7-ketocholesterol-induced apoptosis of U937 cells"'
        self.parser.parseString(s)

        self.assertIsNotNone(self.parser.evidence)

        expected_annotation = {
            CITATION: test_citation_dict,
            ANNOTATIONS: {},
            EVIDENCE: "For instance, during 7-ketocholesterol-induced apoptosis of U937 cells",
        }

        self.assertEqual(expected_annotation, self.parser.get_annotations())

    def test_custom_annotation(self):
        s = [SET_CITATION_TEST, 'SET Custom1 = "Custom1_A"']
        self.parser.parse_lines(s)

        expected_annotation = {
            "Custom1": [Entity(namespace="Custom1", identifier="Custom1_A")],
        }

        self.assertEqual(expected_annotation, self.parser.annotations)

    def test_custom_annotation_list(self):
        s = [SET_CITATION_TEST, 'SET Custom1 = {"Custom1_A","Custom1_B"}']
        self.parser.parse_lines(s)

        expected_annotation = {
            "Custom1": [
                Entity(namespace="Custom1", identifier="Custom1_A"),
                Entity(namespace="Custom1", identifier="Custom1_B"),
            ],
        }

        self.assertEqual(expected_annotation, self.parser.annotations)

        expected_dict = {
            ANNOTATIONS: expected_annotation,
            CITATION: test_citation_dict,
            EVIDENCE: None,
        }

        self.assertEqual(expected_dict, self.parser.get_annotations())

    def test_overwrite_evidence(self):
        s1 = 'SET Evidence = "a"'
        s2 = 'SET Evidence = "b"'

        self.parser.parseString(s1)
        self.parser.parseString(s2)

        self.assertEqual("b", self.parser.evidence)

    def test_unset_evidence(self):
        s1 = 'SET Evidence = "a"'
        s2 = "UNSET Evidence"

        self.parser.parseString(s1)
        self.parser.parseString(s2)

        self.assertEqual({}, self.parser.annotations)

    def test_unset_custom(self):
        statements = [SET_CITATION_TEST, 'SET Custom1 = "Custom1_A"', "UNSET Custom1"]

        self.parser.parse_lines(statements)

        self.assertEqual({}, self.parser.annotations)

    def test_reset_citation(self):
        s1_identifier = str(randint(0, 1e7))
        s1 = 'SET Citation = {{"PubMed","Test Reference 1","{}"}}'.format(s1_identifier)
        s2 = 'SET Evidence = "d"'

        s3_identifier = str(randint(0, 1e7))
        s3 = 'SET Citation = {{"PubMed","Test Reference 2","{}"}}'.format(s3_identifier)
        _test_evidence = n()
        s4 = 'SET Evidence = "{}"'.format(_test_evidence)
        s5 = 'SET Custom1 = "Custom1_A"'
        s6 = 'SET Custom2 = "Custom2_A"'

        statements = [s1, s2, s3, s4, s5, s6]

        self.parser.parse_lines(statements)

        self.assertEqual(_test_evidence, self.parser.evidence)
        self.assertEqual(CITATION_TYPE_PUBMED, self.parser.citation_db)
        self.assertEqual(s3_identifier, self.parser.citation_db_id)

        self.parser.parseString("UNSET {Custom1,Evidence}")
        self.assertNotIn("Custom1", self.parser.annotations)
        self.assertIsNone(self.parser.evidence)
        self.assertIn("Custom2", self.parser.annotations)
        self.assertTrue(self.parser.citation_is_set)

        self.parser.parseString("UNSET ALL")
        self.assertEqual(0, len(self.parser.annotations))
        self.assertFalse(self.parser.citation_is_set)

    def test_set_regex(self):
        v = str(randint(0, 1e5))
        s = [SET_CITATION_TEST, f'SET CustomRegex = "{v}"']
        self.parser.parse_lines(s)

        self.assertEqual(
            [
                Entity(namespace="CustomRegex", identifier=v),
            ],
            self.parser.annotations["CustomRegex"],
        )



# -*- coding: utf-8 -*-

"""Predicate functions for filtering lists of nodes."""

from .activities import has_activity, has_edge_modifier, is_degraded, is_translocated
from .misc import is_isolated_list_abundance, none_of, one_of
from .modifications import (
    has_fragment,
    has_gene_modification,
    has_hgvs,
    has_protein_modification,
    has_variant,
)
from .relations import (
    has_causal_edges,
    has_causal_in_edges,
    has_causal_out_edges,
    has_in_edges,
    has_out_edges,
    is_causal_central,
    is_causal_sink,
    is_causal_source,
    no_causal_edges,
    no_causal_in_edges,
    no_causal_out_edges,
    no_in_edges,
    no_out_edges,
)
from .types import (
    is_abundance,
    is_biological_process,
    is_central_dogma,
    is_complex,
    is_composite,
    is_gene,
    is_list,
    is_mirna,
    is_pathology,
    is_population,
    is_protein,
    is_reaction,
    is_rna,
    is_transcribable,
    not_abundance,
    not_biological_process,
    not_central_dogma,
    not_complex,
    not_composite,
    not_gene,
    not_list,
    not_mirna,
    not_pathology,
    not_population,
    not_protein,
    not_reaction,
    not_rna,
    not_transcribable,
)
from .utils import (
    concatenate_node_predicates,
    false_node_predicate,
    invert_node_predicate,
    node_predicate,
    true_node_predicate,
)

__all__ = [k for k in locals() if not k.startswith("_")]



# -*- coding: utf-8 -*-

"""Utilities for BEL repositories."""

import json
import logging
import os
import sys
import time
from dataclasses import dataclass, field
from itertools import chain
from typing import Any, Iterable, Mapping, Optional, Set, TextIO, Tuple, Union

import click
import pandas as pd
from tqdm.autonotebook import tqdm

from .cli import (
    connection_option,
    host_option,
    password_option,
    user_option,
    verbose_option,
)
from .constants import CITATION
from .io import from_bel_script, to_bel_commons, to_indra_statements
from .io.api import dump, load
from .manager import Manager
from .manager.citation_utils import enrich_pubmed_citations
from .struct import BELGraph
from .struct.operations import union
from .version import get_version

__all__ = [
    "BELMetadata",
    "BELRepository",
    "append_click_group",
]

logger = logging.getLogger(__name__)

private_option = click.option("--private", is_flag=True)

OUTPUT_KWARGS = {
    "nodelink.json": dict(indent=2, sort_keys=True),
    "cx.json": dict(indent=2, sort_keys=True),
    "jgif.json": dict(indent=2, sort_keys=True),
}


@dataclass
class BELMetadata:
    """A container for BEL document metadata."""

    name: Optional[str] = None
    version: Optional[str] = None
    description: Optional[str] = None
    authors: Optional[str] = None
    contact: Optional[str] = None
    license: Optional[str] = None
    copyright: Optional[str] = None
    disclaimer: Optional[str] = None

    def new(self) -> BELGraph:
        """Generate a new BEL graph with the given metadata."""
        graph = BELGraph()
        self.update(graph)
        return graph

    def update(self, graph: BELGraph) -> None:
        """Update the BEL graph's metadata."""
        if self.name:
            graph.name = self.name
        if self.version:
            graph.version = self.version
        if self.authors:
            graph.authors = self.authors
        if self.description:
            graph.description = self.description
        if self.contact:
            graph.contact = self.contact
        if self.license:
            graph.licenses = self.license
        if self.copyright:
            graph.copyright = self.copyright
        if self.disclaimer:
            graph.disclaimer = self.disclaimer


@dataclass
class BELRepository:
    """A container for a BEL repository."""

    directory: str
    output_directory: Optional[str] = None

    bel_cache_name: str = "_cache.bel"
    metadata: Optional[BELMetadata] = None
    formats: Tuple[str, ...] = ("pickle", "nodelink.json")

    #: Must include {file_name} and {extension}
    cache_fmt: str = "{file_name}.{extension}"
    global_summary_ext: str = "summary.tsv"
    warnings_ext: str = "warnings.tsv"

    #: Arguments passed to :func:`pybel.from_path` during compilation
    from_path_kwargs: Mapping[str, Any] = field(default_factory=dict)

    #: The location where the summary DataFrame will be output as a TSV.
    bel_summary_path: str = field(init=False)

    def __post_init__(self) -> None:  # noqa: D105
        if self.output_directory is None:
            self.output_directory = self.directory

        self.bel_summary_path = self._build_cache_ext_path(
            root=self.output_directory,
            file_name=self.bel_cache_name,
            extension=self.global_summary_ext.lstrip("."),
        )

    def _get_global_cache_path_by_extension(self, extension: str) -> str:
        return self._build_cache_ext_path(self.output_directory, self.bel_cache_name, extension)

    def _build_warnings_path(self, root: str, file_name: str) -> str:
        return self._build_cache_ext_path(root, file_name, self.warnings_ext.lstrip("."))

    def _build_summary_path(self, root: str, file_name: str) -> str:
        return self._build_cache_ext_path(root, file_name, "summary.json")

    def _build_cache_ext_path(self, root: str, file_name: str, extension: str) -> str:
        return os.path.join(
            root,
            self.cache_fmt.format(file_name=file_name, extension=extension.lstrip(".")),
        )

    def walk(self) -> Iterable[Tuple[str, Iterable[str], Iterable[str]]]:
        """Recursively walk this directory."""
        return os.walk(self.directory)

    def iterate_bel(self) -> Iterable[Tuple[str, str]]:
        """Yield all paths to BEL documents."""
        for root, _dirs, file_names in self.walk():
            for file_name in sorted(file_names):
                if not file_name.startswith("_") and file_name.endswith(".bel"):
                    yield root, file_name

    def clear_global_cache(self) -> None:
        """Clear the global cache."""
        self._remove_root_file_name(self.output_directory, self.bel_cache_name)

    def clear_local_caches(self) -> None:
        """Clear all caches of BEL documents in the repository."""
        for root, file_name in self.iterate_bel():
            self._remove_root_file_name(root, file_name)

    def clear_local_warned(self) -> None:
        """Clear caches for BEL documents with errors."""
        for root, file_name in self.iterate_bel():
            if self._has_warnings(root, file_name):
                self._remove_root_file_name(root, file_name)

    def _has_warnings(self, root: str, file_name: str) -> bool:
        return os.path.exists(self._build_warnings_path(root, file_name))

    def _remove_root_file_name(self, root: str, file_name: str) -> None:
        for _, path in self._iterate_extension_path(root, file_name):
            if os.path.exists(path):
                os.remove(path)

    def _iterate_extension_path(self, root: str, file_name: str) -> Iterable[Tuple[str, str]]:
        for extension in self.formats:
            yield extension, self._build_cache_ext_path(root, file_name, extension)

    def _import_local(self, root: str, file_name: str) -> Optional[BELGraph]:
        for _, path in self._iterate_extension_path(root, file_name):
            if os.path.exists(path):
                return load(path)

        return None

    def _import_global(self) -> Optional[BELGraph]:
        return self._import_local(self.output_directory, self.bel_cache_name)

    def _export_local(self, graph: BELGraph, root: str, file_name: str) -> None:
        for extension, path in self._iterate_extension_path(root, file_name):
            kwargs = OUTPUT_KWARGS.get(extension, {})
            dump(graph, path, **kwargs)

        with open(self._build_summary_path(root, file_name), "w") as file:
            json.dump(graph.summarize.dict(), file, indent=2)

        if graph.warnings:
            logger.info(f" - {graph.number_of_warnings()} warnings")
            warnings_path = self._build_warnings_path(root, file_name)
            warnings_df = pd.DataFrame(
                [
                    (
                        exc.line_number,
                        exc.position,
                        exc.line,
                        exc.__class__.__name__,
                        str(exc),
                    )
                    for _, exc, _ in graph.warnings
                ],
                columns=["Line Number", "Position", "Line", "Error", "Message"],
            )
            warnings_df.to_csv(warnings_path, sep="\t", index=False)

    def _export_global(self, graph: BELGraph) -> None:
        self._export_local(graph, self.output_directory, self.bel_cache_name)

    def get_graph(
        self,
        manager: Optional[Manager] = None,
        use_cached: bool = True,
        use_tqdm: bool = False,
        tqdm_kwargs: Optional[Mapping[str, Any]] = None,
        from_path_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> BELGraph:
        """Get a combine graph."""
        if use_cached:
            graph = self._import_global()
            if graph is not None:
                return graph

        graphs = self.get_graphs(
            manager=manager,
            use_tqdm=use_tqdm,
            tqdm_kwargs=tqdm_kwargs,
            from_path_kwargs=from_path_kwargs,
        )
        graph = union(graphs.values())

        if self.metadata is not None:
            self.metadata.update(graph)

        self._get_summary_df_from_graphs(graphs)
        self._export_global(graph)

        return graph

    def get_indra_statements(self, **kwargs):
        """Get INDRA statements for all graphs.

        :rtype: List[indra.statements.Statement]
        """
        return list(chain.from_iterable(to_indra_statements(graph) for graph in self.get_graphs(**kwargs).values()))

    def get_graphs(
        self,
        manager: Optional[Manager] = None,
        use_cached: bool = True,
        use_tqdm: bool = False,
        tqdm_kwargs: Optional[Mapping[str, Any]] = None,
        from_path_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> Mapping[str, BELGraph]:
        """Get a mapping of all graphs' paths to their compiled BEL graphs."""
        if manager is None:
            manager = Manager()

        paths = self.iterate_bel()
        if use_tqdm:
            paths = tqdm(list(paths), **(tqdm_kwargs or {}))

        rv = {}
        for root, file_name in paths:
            path = os.path.join(root, file_name)

            if use_cached:
                graph = self._import_local(root, file_name)
                if graph is not None:
                    rv[path] = graph
                    continue

            _from_path_kwargs = from_path_kwargs or {}
            _from_path_kwargs.update(self.from_path_kwargs)

            try:
                graph = rv[path] = from_bel_script(path, manager=manager, **_from_path_kwargs)
                graph.path = os.path.relpath(os.path.join(root, file_name), self.directory)
            except Exception as exc:
                logger.warning(f"problem with {path}: {exc}")
                continue

            enrich_pubmed_citations(graph=graph, manager=manager)
            self._export_local(graph, root, file_name)

        return rv

    def get_summary_df(
        self,
        manager: Optional[Manager] = None,
        use_cached: bool = False,
        use_tqdm: bool = False,
        tqdm_kwargs: Optional[Mapping[str, Any]] = None,
        from_path_kwargs: Optional[Mapping[str, Any]] = None,
        save: Union[bool, str, TextIO] = True,
    ) -> pd.DataFrame:
        """Get a pandas DataFrame summarizing the contents of all graphs in the repository."""
        graphs = self.get_graphs(
            manager=manager,
            use_cached=use_cached,
            use_tqdm=use_tqdm,
            tqdm_kwargs=tqdm_kwargs,
            from_path_kwargs=from_path_kwargs,
        )
        return self._get_summary_df_from_graphs(graphs, save=save)

    def _get_summary_df_from_graphs(self, graphs, save: Union[str, bool, TextIO] = True):
        summary_dicts = {
            os.path.relpath(path, self.directory): graph.summarize.dict() for path, graph in graphs.items()
        }

        df = pd.DataFrame.from_dict(summary_dicts, orient="index")

        if isinstance(save, str):
            df.to_csv(save, sep="\t")
        elif save:
            df.to_csv(self.bel_summary_path, sep="\t")

        return df

    def build_cli(self):  # noqa: D202
        """Build a command line interface."""

        @click.group(help=f"Tools for the BEL repository at {self.directory} using PyBEL v{get_version()}")
        @click.pass_context
        def main(ctx):
            """Group the commands."""
            ctx.obj = self

        append_click_group(main)
        return main

    def get_extensions(self, root: str, file_name: str) -> Set[str]:
        """Get all compiled files for the given BEL."""
        # TODO check that this is a valid BEL path!
        return {extension for extension, path in self._iterate_extension_path(root, file_name) if os.path.exists(path)}

    def _get_global_caches(self):
        return self.get_extensions(self.output_directory, self.bel_cache_name)

    def _iterate_citations(self, **kwargs) -> Iterable[Tuple[str, str]]:
        """List all citations in documents in this repository."""
        for _, _, data in self.get_graph(**kwargs).edges(data=True):
            citation = data.get(CITATION)
            if citation is not None:
                yield citation.namespace, citation.identifier


def _write_caches(bel_repository: BELRepository, root: str, file_name: str):
    extensions = ", ".join(sorted(bel_repository.get_extensions(root, file_name)))
    has_warnings = os.path.exists(bel_repository._build_warnings_path(root, file_name))

    try:
        with open(bel_repository._build_summary_path(root, file_name)) as file:
            summary = json.load(file)
    except FileNotFoundError:
        summary = None

    if extensions and has_warnings:
        s = click.style("✘️ ", fg="red")
    elif extensions and not has_warnings:
        s = click.style("✔︎ ", fg="green")
    else:
        s = click.style("? ", fg="yellow", bold=True)

    path = os.path.join(root, file_name)
    s += os.path.relpath(path, bel_repository.directory)

    if extensions:
        s += click.style(f" ({extensions})", fg="green")

    if summary:
        s += click.style(
            f' ({summary["Number of Nodes"]} nodes, {summary["Number of Edges"]} edges)',
            fg="blue",
        )

    click.echo(s)


def append_click_group(group: click.Group) -> None:  # noqa: D202, C901
    """Append a :py:class:`click.Group`."""

    @group.command()
    @click.pass_obj
    def ls(bel_repository: BELRepository):
        """List the contents of the repository."""
        global_caches = bel_repository._get_global_caches()
        if global_caches:
            click.secho("Global Cache", fg="red", bold=True)
            _write_caches(
                bel_repository,
                bel_repository.output_directory,
                bel_repository.bel_cache_name,
            )
            click.secho("Local Caches", fg="red", bold=True)

        for root, file_name in bel_repository.iterate_bel():
            _write_caches(bel_repository, root, file_name)

    @group.command()
    @click.pass_obj
    def citations(repository: BELRepository):
        """List citations in the repository."""
        for database, reference in sorted(set(repository._iterate_citations(use_tqdm=True)), key=lambda x: int(x[1])):
            click.echo(f"{database}\t{reference}")

    @group.command()
    @host_option
    @user_option
    @password_option
    @click.option("-s", "--sleep", type=int, default=3, help="Seconds to sleep between sending")
    @private_option
    @click.pass_obj
    def upload_separate(
        repository: BELRepository,
        host: str,
        user: str,
        password: str,
        sleep: int,
        private: bool,
    ):
        """Upload all to BEL Commons."""
        it = tqdm(repository.get_graphs().items())
        for name, graph in it:
            res = to_bel_commons(graph, host=host, user=user, password=password, public=not private)
            res_json = res.json()
            task_id = res_json.get("task_id")
            if task_id is not None:
                it.write(f"task:{task_id} - {name}")
                it.write(f'see: {host.rstrip("/")}/api/task/{task_id}')
                time.sleep(sleep)
            else:
                it.write(f"problem with {name}: {res_json}")

    @group.command()
    @host_option
    @user_option
    @password_option
    @private_option
    @click.pass_obj
    def upload_combine(repository: BELRepository, host: str, user: str, password: str, private: bool):
        """Upload the combine graph."""
        graph = repository.get_graph()
        res = to_bel_commons(graph, host=host, user=user, password=password, public=not private)
        res_json = res.json()
        task_id = res_json.get("task_id")
        if task_id is not None:
            click.echo(f"task:{task_id} - {graph}")
            click.echo(f'see: {host.rstrip("/")}/api/task/{task_id}')
        else:
            click.echo(f"problem with {graph.name}: {res_json}")

    @group.command()
    @click.confirmation_option()
    @click.pass_obj
    def uncache(bel_repository: BELRepository):
        """Clear the cached data for the repository."""
        bel_repository.clear_global_cache()
        bel_repository.clear_local_caches()

    @group.command()
    @click.confirmation_option()
    @click.pass_obj
    def uncache_global(bel_repository: BELRepository):
        """Clear the cached data for the repository."""
        bel_repository.clear_global_cache()

    @group.command()
    @click.confirmation_option()
    @click.pass_obj
    def uncache_local(bel_repository: BELRepository):
        """Clear the cached data for the repository."""
        bel_repository.clear_local_caches()

    @group.command()
    @click.confirmation_option()
    @click.pass_obj
    def uncache_warned(bel_repository: BELRepository):
        """Clear the cached data for the documents that have warnings."""
        bel_repository.clear_local_warned()

    @group.command()
    @connection_option
    @click.option("-r", "--reload", is_flag=True)
    @click.option("--no-tqdm", is_flag=True)
    @verbose_option
    @click.pass_obj
    def compile(bel_repository: BELRepository, connection: str, reload: bool, no_tqdm: bool):
        """Summarize the repository."""
        if reload:
            bel_repository.clear_global_cache()
            bel_repository.clear_local_caches()

        manager = Manager(connection=connection)
        graph = bel_repository.get_graph(
            manager=manager,
            use_cached=(not reload),
            use_tqdm=(not no_tqdm),
            tqdm_kwargs=dict(
                desc="Loading BEL",
                leave=False,
            ),
            from_path_kwargs=dict(
                use_tqdm=(not no_tqdm),
                tqdm_kwargs=dict(
                    leave=False,
                ),
            ),
        )
        click.echo(graph.summarize.str())

    @group.command()
    @click.argument("file", type=click.File("w"))
    @click.pass_obj
    def html(bel_repository: BELRepository, file: TextIO):
        """Output an HTML summary."""
        graph = bel_repository.get_graph()
        try:
            from pybel_tools.assembler.html import to_html_file
        except ImportError:
            click.secho("pybel_tools.assembler.html is not available", fg="red")
            sys.exit(1)
        else:
            to_html_file(graph, file)


@click.group()
@click.version_option()
@click.option(
    "-d",
    "--directory",
    default=os.getcwd(),
    type=click.Path(file_okay=False, dir_okay=True, exists=True),
    help="Defaults to current working directory",
)
@click.pass_context
def main(ctx, directory: str):
    """Command line interface for bel-repository."""
    ctx.obj = BELRepository(directory=directory)


append_click_group(main)

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""Functions for predicates for edge data from BEL graphs."""

from typing import Iterable, Mapping

from .edge_filters import invert_edge_predicate
from .edge_predicates import (
    edge_predicate,
    has_authors,
    has_pubmed,
    true_edge_predicate,
)
from .typing import EdgePredicate
from ..graph import BELGraph
from ...constants import (
    ANNOTATIONS,
    CAUSAL_RELATIONS,
    CITATION,
    CITATION_AUTHORS,
    IDENTIFIER,
    RELATION,
)
from ...dsl import BaseEntity
from ...typing import EdgeData, Strings

__all__ = [
    "build_pmid_exclusion_filter",
    "build_annotation_dict_all_filter",
    "build_annotation_dict_any_filter",
    "build_upstream_edge_predicate",
    "build_downstream_edge_predicate",
    "build_relation_predicate",
    "build_pmid_inclusion_filter",
    "build_pmid_exclusion_filter",
    "build_author_inclusion_filter",
]


def _annotation_dict_all_filter(edge_data: EdgeData, query: Mapping[str, Iterable[str]]) -> bool:
    """Match edges with the given dictionary as a sub-dictionary.

    :param query: The annotation query dict to match
    """
    annotations = edge_data.get(ANNOTATIONS)

    if annotations is None:
        return False

    for key, values in query.items():
        ak = annotations.get(key)

        if ak is None:
            return False

        for value in values:
            if value not in ak:
                return False

    return True


def build_annotation_dict_all_filter(annotations: Mapping[str, Iterable[str]]) -> EdgePredicate:
    """Build an edge predicate for edges whose annotations are super-dictionaries of the given dictionary.

    If no annotations are given, will always evaluate to true.

    :param annotations: The annotation query dict to match
    """
    if not annotations:
        return true_edge_predicate

    @edge_predicate
    def annotation_dict_all_filter(edge_data: EdgeData) -> bool:
        """Check if the all of the annotations in the enclosed query match."""
        return _annotation_dict_all_filter(edge_data, query=annotations)

    return annotation_dict_all_filter


def _annotation_dict_any_filter(edge_data: EdgeData, query: Mapping[str, Iterable[str]]) -> bool:
    """Match edges with the given dictionary as a sub-dictionary.

    :param query: The annotation query dict to match
    """
    annotations = edge_data.get(ANNOTATIONS)

    if annotations is None:
        return False

    return any(key in annotations and value in annotations[key] for key, values in query.items() for value in values)


def build_annotation_dict_any_filter(annotations: Mapping[str, Iterable[str]]) -> EdgePredicate:
    """Build an edge predicate that passes for edges whose data dictionaries match the given dictionary.

    If the given dictionary is empty, will always evaluate to true.

    :param annotations: The annotation query dict to match
    """
    if not annotations:
        return true_edge_predicate

    @edge_predicate
    def annotation_dict_any_filter(edge_data: EdgeData) -> bool:
        """Check if the any of the annotations in the enclosed query match."""
        return _annotation_dict_any_filter(edge_data, query=annotations)

    return annotation_dict_any_filter


def build_upstream_edge_predicate(nodes: Iterable[BaseEntity]) -> EdgePredicate:
    """Build an edge predicate that pass for relations for which one of the given nodes is the object."""
    nodes = set(nodes)

    def upstream_filter(graph: BELGraph, u: BaseEntity, v: BaseEntity, k: str) -> bool:
        """Pass for relations for which one of the given nodes is the object."""
        return v in nodes and graph[u][v][k][RELATION] in CAUSAL_RELATIONS

    return upstream_filter


def build_downstream_edge_predicate(nodes: Iterable[BaseEntity]) -> EdgePredicate:
    """Build an edge predicate that passes for edges for which one of the given nodes is the subject."""
    nodes = set(nodes)

    def downstream_filter(graph: BELGraph, u: BaseEntity, v: BaseEntity, k: str) -> bool:
        """Pass for relations for which one of the given nodes is the subject."""
        return u in nodes and graph[u][v][k][RELATION] in CAUSAL_RELATIONS

    return downstream_filter


def build_relation_predicate(relations: Strings) -> EdgePredicate:
    """Build an edge predicate that passes for edges with the given relation."""
    if isinstance(relations, str):

        @edge_predicate
        def relation_predicate(edge_data: EdgeData) -> bool:
            """Pass for relations matching the enclosed value."""
            return edge_data[RELATION] == relations

    elif isinstance(relations, Iterable):
        relation_set = set(relations)

        @edge_predicate
        def relation_predicate(edge_data: EdgeData) -> bool:
            """Pass for relations matching the enclosed values."""
            return edge_data[RELATION] in relation_set

    else:
        raise TypeError

    return relation_predicate


def build_pmid_inclusion_filter(pmids: Strings) -> EdgePredicate:
    """Build an edge predicate that passes for edges with citations from the given PubMed identifier(s).

    :param pmids: A PubMed identifier or list of PubMed identifiers to filter for
    """
    if isinstance(pmids, str):

        @edge_predicate
        def pmid_inclusion_filter(edge_data: EdgeData) -> bool:
            """Pass for edges with PubMed citations matching the contained PubMed identifier."""
            return has_pubmed(edge_data) and edge_data[CITATION].identifier == pmids

    elif isinstance(pmids, Iterable):
        pmids = set(pmids)

        @edge_predicate
        def pmid_inclusion_filter(edge_data: EdgeData) -> bool:
            """Pass for edges with PubMed citations matching one of the contained PubMed identifiers."""
            return has_pubmed(edge_data) and edge_data[CITATION].identifier in pmids

    else:
        raise TypeError

    return pmid_inclusion_filter


def build_pmid_exclusion_filter(pmids: Strings) -> EdgePredicate:
    """Fail for edges with citations whose references are one of the given PubMed identifiers.

    :param pmids: A PubMed identifier or list of PubMed identifiers to filter against
    """
    return invert_edge_predicate(build_pmid_inclusion_filter(pmids))


def build_author_inclusion_filter(authors: Strings) -> EdgePredicate:
    """Build an edge predicate that passes for edges with citations written by the given author(s)."""
    if isinstance(authors, str):

        @edge_predicate
        def author_filter(edge_data: EdgeData) -> bool:
            """Pass for edges with citations with an author that matches the contained author."""
            return has_authors(edge_data) and authors in edge_data[CITATION][CITATION_AUTHORS]

    elif isinstance(authors, Iterable):
        authors = set(authors)

        @edge_predicate
        def author_filter(edge_data: EdgeData) -> bool:
            """Pass for edges with citations with an author that matches one or more of the contained authors."""
            return has_authors(edge_data) and any(author in edge_data[CITATION][CITATION_AUTHORS] for author in authors)

    else:
        raise TypeError

    return author_filter



# -*- coding: utf-8 -*-

"""Conversion functions for BEL graphs with JGIF JSON.

The JSON Graph Interchange Format (JGIF) is `specified <http://jsongraphformat.info/>`_ similarly to the Node-Link
JSON. Interchange with this format provides compatibilty with other software and repositories, such as the
`Causal Biological Network Database <http://causalbionet.com/>`_.
"""

import gzip
import json
import logging
import re
from collections import defaultdict
from operator import methodcaller
from typing import Any, Mapping, Optional, TextIO, Union

import requests
from networkx.utils import open_file
from pyparsing import ParseException

from .. import constants as pc
from ..constants import (
    ANNOTATIONS,
    CITATION,
    EVIDENCE,
    GRAPH_ANNOTATION_URL,
    GRAPH_NAMESPACE_URL,
    METADATA_AUTHORS,
    METADATA_CONTACT,
    METADATA_INSERT_KEYS,
    METADATA_LICENSES,
    RELATION,
    UNQUALIFIED_EDGES,
)
from ..exceptions import NakedNameWarning, UndefinedNamespaceWarning
from ..parser import BELParser
from ..struct import BELGraph
from ..version import get_version

__all__ = [
    "from_cbn_jgif",
    "from_cbn_jgif_file",
    "from_jgif",
    "from_jgif_file",
    "from_jgif_gz",
    "from_jgif_jsons",
    "to_jgif",
    "to_jgif_file",
    "to_jgif_gz",
    "to_jgif_jsons",
    "post_jgif",
]

logger = logging.getLogger(__name__)

annotation_map = {
    "tissue": "Tissue",
    "disease": "Disease",
    "species_common_name": "Species",
    "cell": "Cell",
}

species_map = {
    "human": "9606",
    "rat": "10116",
    "mouse": "10090",
}

placeholder_evidence = (
    "This Network edge has no supporting evidence.  Please add real evidence to this edge prior to deleting."
)

EXPERIMENT_CONTEXT = "experiment_context"


def map_cbn(d):
    """Pre-processes the JSON from the CBN.

    - removes statements without evidence, or with placeholder evidence

    :param dict d: Raw JGIF from the CBN
    :return: Preprocessed JGIF
    :rtype: dict
    """
    for i, edge in enumerate(d["graph"]["edges"]):
        if "metadata" not in edge:
            continue

        if "evidences" not in edge["metadata"]:
            continue

        for j, evidence in enumerate(edge["metadata"]["evidences"]):
            if EXPERIMENT_CONTEXT not in evidence:
                continue

            # ctx = {k.strip().lower(): v.strip() for k, v in evidence[EXPERIMENT_CONTEXT].items() if v.strip()}

            new_context = {}

            for key, value in evidence[EXPERIMENT_CONTEXT].items():
                if not value:
                    logger.debug("key %s without value", key)
                    continue

                value = value.strip()

                if not value:
                    logger.debug("key %s without value", key)
                    continue

                key = key.strip().lower()

                if key == "species_common_name":
                    new_context["Species"] = species_map[value.lower()]
                elif key in annotation_map:
                    new_context[annotation_map[key]] = value
                else:
                    new_context[key] = value

            """
            for k, v in annotation_map.items():
                if k not in ctx:
                    continue

                d['graph']['edges'][i]['metadata']['evidences'][j][EXPERIMENT_CONTEXT][v] = ctx[k]
                del d['graph']['edges'][i]['metadata']['evidences'][j][EXPERIMENT_CONTEXT][k]

            if 'species_common_name' in ctx:
                species_name = ctx['species_common_name'].strip().lower()
                d['graph']['edges'][i]['metadata']['evidences'][j][EXPERIMENT_CONTEXT]['Species'] = species_map[
                    species_name]
                del d['graph']['edges'][i]['metadata']['evidences'][j][EXPERIMENT_CONTEXT][
                    'species_common_name']
            """

            # TODO can this be replaced with edge as well?
            d["graph"]["edges"][i]["metadata"]["evidences"][j][EXPERIMENT_CONTEXT] = new_context

    return d


NAMESPACE_URLS = {
    "TAX": "https://arty.scai.fraunhofer.de/artifactory/bel/namespace/ncbi-taxonomy/ncbi-taxonomy-20200322.belns",
    "HGNC": "https://arty.scai.fraunhofer.de/artifactory/bel/namespace/hgnc-human-genes/hgnc-human-genes-20150601.belns",
    "GOBP": "https://arty.scai.fraunhofer.de/artifactory/bel/namespace/go-biological-process/go-biological-process-20150601.belns",
    "SFAM": "https://arty.scai.fraunhofer.de/artifactory/bel/namespace/selventa-protein-families/selventa-protein-families-20150601.belns",
    "GOCC": "https://arty.scai.fraunhofer.de/artifactory/bel/namespace/go-cellular-component/go-cellular-component-20170511.belns",
    "MESHPP": "https://arty.scai.fraunhofer.de/artifactory/bel/namespace/mesh-processes/mesh-processes-20150601.belns",
    "MGI": "https://arty.scai.fraunhofer.de/artifactory/bel/namespace/mgi-mouse-genes/mgi-mouse-genes-20150601.belns",
    "RGD": "https://arty.scai.fraunhofer.de/artifactory/bel/namespace/rgd-rat-genes/rgd-rat-genes-20150601.belns",
    "CHEBI": "https://arty.scai.fraunhofer.de/artifactory/bel/namespace/chebi/chebi-20150601.belns",
    "SCHEM": "https://arty.scai.fraunhofer.de/artifactory/bel/namespace/selventa-legacy-chemicals/selventa-legacy-chemicals-20150601.belns",
    "EGID": "https://arty.scai.fraunhofer.de/artifactory/bel/namespace/entrez-gene-ids/entrez-gene-ids-20150601.belns",
    "MESHD": "https://arty.scai.fraunhofer.de/artifactory/bel/namespace/mesh-diseases/mesh-diseases-20150601.belns",
    "SDIS": "https://arty.scai.fraunhofer.de/artifactory/bel/namespace/selventa-legacy-diseases/selventa-legacy-diseases-20150601.belns",
    "SCOMP": "https://arty.scai.fraunhofer.de/artifactory/bel/namespace/selventa-named-complexes/selventa-named-complexes-20150601.belns",
    "MESHC": "https://arty.scai.fraunhofer.de/artifactory/bel/namespace/mesh-chemicals/mesh-chemicals-20170511.belns",
    "GOBPID": "https://arty.scai.fraunhofer.de/artifactory/bel/namespace/go-biological-process-ids/go-biological-process-ids-20150601.belns",
    "GOCCID": "https://arty.scai.fraunhofer.de/artifactory/bel/namespace/go-cellular-component-ids/go-cellular-component-ids-20150601.belns",
    "MESHCS": "https://arty.scai.fraunhofer.de/artifactory/bel/namespace/mesh-cell-structures/mesh-cell-structures-20150601.belns",
}

ANNOTATION_URLS = {
    "Cell": "https://arty.scai.fraunhofer.de/artifactory/bel/annotation/cell-line/cell-line-20150601.belanno",
    "Disease": "https://arty.scai.fraunhofer.de/artifactory/bel/annotation/disease/disease-20150601.belanno",
    "Species": "https://arty.scai.fraunhofer.de/artifactory/bel/annotation/species-taxonomy-id/species-taxonomy-id-20170511.belanno",
    "Tissue": "https://arty.scai.fraunhofer.de/artifactory/bel/annotation/mesh-anatomy/mesh-anatomy-20150601.belanno",
}

NAMESPACE_TO_PATTERN = {
    namespace: re.compile(r".*") for namespace in (set(NAMESPACE_URLS) | {"GO", "MESH"})  # don't validate anything
}


@open_file(0, mode="r")
def from_cbn_jgif_file(path: Union[str, TextIO]) -> BELGraph:
    """Build a graph from a file containing the CBN variant of JGIF.

    :param path: A path or file-like
    """
    return from_cbn_jgif(json.load(path))


def from_cbn_jgif(graph_jgif_dict):
    """Build a BEL graph from CBN JGIF.

    Map the JGIF used by the Causal Biological Network Database to standard namespace and annotations, then
    builds a BEL graph using :func:`pybel.from_jgif`.

    :param dict graph_jgif_dict: The JSON object representing the graph in JGIF format
    :rtype: BELGraph

    Example:
    .. code-block:: python

        import requests
        from pybel import from_cbn_jgif
        apoptosis_url = 'http://causalbionet.com/Networks/GetJSONGraphFile?networkId=810385422'
        graph_jgif_dict = requests.get(apoptosis_url).json()
        graph = from_cbn_jgif(graph_jgif_dict)

    .. warning::

        Handling the annotations is not yet supported, since the CBN documents do not refer to the resources used
        to create them. This may be added in the future, but the annotations must be stripped from the graph
        before uploading to the network store using :func:`pybel.struct.mutation.strip_annotations`.

    """
    graph_jgif_dict = map_cbn(graph_jgif_dict)

    graph_jgif_dict["graph"][GRAPH_NAMESPACE_URL] = NAMESPACE_URLS
    graph_jgif_dict["graph"][GRAPH_ANNOTATION_URL] = ANNOTATION_URLS
    graph_jgif_dict["graph"]["metadata"].update(
        {
            METADATA_AUTHORS: "Causal Biological Networks Database",
            METADATA_LICENSES: """
        Please cite:

        - www.causalbionet.com
        - https://bionet.sbvimprover.com

        as well as any relevant publications.

        The sbv IMPROVER project, the website and the Symposia are part of a collaborative project
        designed to enable scientists to learn about and contribute to the development of a new crowd
        sourcing method for verification of scientific data and results. The current challenges, website
        and biological network models were developed and are maintained as part of a collaboration among
        Selventa, OrangeBus and ADS. The project is led and funded by Philip Morris International. For more
        information on the focus of Philip Morris International’s research, please visit www.pmi.com.
        """.replace(
                "\n", "\t"
            ),
            METADATA_CONTACT: "CausalBiologicalNetworks.RD@pmi.com",
        }
    )

    graph = from_jgif(graph_jgif_dict)
    return graph


def from_jgif(graph_jgif_dict, parser_kwargs: Optional[Mapping[str, Any]] = None):  # noqa:C901
    """Build a BEL graph from a JGIF JSON object.

    :param dict graph_jgif_dict: The JSON object representing the graph in JGIF format
    :rtype: BELGraph
    """
    graph = BELGraph()

    root = graph_jgif_dict["graph"]

    if "label" in root:
        graph.name = root["label"]

    if "metadata" in root:
        metadata = root["metadata"]

        for key in METADATA_INSERT_KEYS:
            if key in metadata:
                graph.document[key] = metadata[key]

    for k in (GRAPH_ANNOTATION_URL, GRAPH_NAMESPACE_URL):
        if k in root:
            graph.graph[k] = root[k]

    parser = BELParser(graph, namespace_to_pattern=NAMESPACE_TO_PATTERN)
    parser.bel_term.addParseAction(parser.handle_term)

    for node in root["nodes"]:
        node_label = node.get("label")

        if node_label is None:
            logger.warning("node missing label: %s", node)
            continue

        try:
            parser.bel_term.parseString(node_label)
        except NakedNameWarning as e:
            logger.info("Naked name: %s", e)
        except UndefinedNamespaceWarning as e:
            logger.info("Undefined namespace: %s", e)
        except ParseException:
            logger.info("Parse exception for %s", node_label)

    for i, edge in enumerate(root["edges"]):
        relation = edge.get("relation")
        if relation is None:
            logger.warning("no relation for edge: %s", edge)

        if relation in {"actsIn", "translocates"}:
            continue  # don't need legacy BEL format

        edge_metadata = edge.get("metadata")
        if edge_metadata is None:
            logger.warning("no metadata for edge: %s", edge)
            continue

        bel_statement = edge.get("label")
        if bel_statement is None:
            logger.debug("No BEL statement for edge %s", edge)

        evidences = edge_metadata.get("evidences")

        if relation in UNQUALIFIED_EDGES:
            pass  # FIXME?

        else:
            if not evidences:  # is none or is empty list
                logger.debug("No evidence for edge %s", edge)
                continue

            for evidence in evidences:
                citation = evidence.get("citation")

                if not citation:
                    continue

                if "type" not in citation or "id" not in citation:
                    continue

                summary_text = evidence["summary_text"].strip()

                if not summary_text or summary_text == placeholder_evidence:
                    continue

                parser.control_parser.clear()
                citation_namespace = citation["type"].lower().strip()
                citation_namespace = pc.CITATION_NORMALIZER.get(citation_namespace, citation_namespace)
                parser.control_parser.citation_db = citation_namespace
                parser.control_parser.citation_db_id = citation["id"].strip()
                parser.control_parser.evidence = summary_text
                annotations = parser.graph._clean_annotations(evidence[EXPERIMENT_CONTEXT])
                parser.control_parser.annotations.update(annotations)

                try:
                    parser.parseString(bel_statement, line_number=i)
                except Exception as e:
                    logger.warning("JGIF relation parse error: %s for %s", e, bel_statement)

    return graph


@open_file(0, mode="r")
def from_jgif_file(path: Union[str, TextIO]) -> BELGraph:
    """Build a graph from the JGIF JSON contained in the given file.

    :param path: A path or file-like
    """
    return from_jgif(json.load(path))


def from_jgif_gz(path: str) -> BELGraph:
    """Read a graph as JGIF JSON from a gzip file."""
    with gzip.open(path, "rt") as file:
        return from_jgif(json.load(file))


def from_jgif_jsons(graph_json_str: str) -> BELGraph:
    """Read a BEL graph from a JGIF JSON string."""
    return from_jgif(json.loads(graph_json_str))


def to_jgif(graph):
    """Build a JGIF dictionary from a BEL graph.

    :param pybel.BELGraph graph: A BEL graph
    :return: A JGIF dictionary
    :rtype: dict

    .. warning::

        Untested! This format is not general purpose and is therefore time is not heavily invested. If you want to
        use Cytoscape.js, we suggest using :func:`pybel.to_cx` instead.

    The example below shows how to output a BEL graph as a JGIF dictionary.

    .. code-block:: python

        import os
        from pybel.examples import sialic_acid_graph
        graph_jgif_json = pybel.to_jgif(sialic_acid_graph)

    If you want to write the graph directly to a file as JGIF, see func:`to_jgif_file`.
    """
    u_v_r_bel = {}

    nodes_entry = []
    edges_entry = []

    for node in sorted(graph, key=methodcaller("as_bel")):
        nodes_entry.append(
            {
                "id": node.md5,
                "label": node.as_bel(),
                "bel_function_type": node.function,
            }
        )

    for u, v in graph.edges():
        relation_evidences = defaultdict(list)

        for key, data in graph[u][v].items():
            if (u, v, data[RELATION]) not in u_v_r_bel:
                u_v_r_bel[u, v, data[RELATION]] = graph.edge_to_bel(u, v, edge_data=data)

            bel = u_v_r_bel[u, v, data[RELATION]]

            evidence_dict = {
                "bel_statement": bel,
                "key": key,
            }

            if ANNOTATIONS in data:
                evidence_dict["experiment_context"] = data[ANNOTATIONS]

            if EVIDENCE in data:
                evidence_dict["summary_text"] = data[EVIDENCE]

            if CITATION in data:
                evidence_dict["citation"] = data[CITATION]

            relation_evidences[data[RELATION]].append(evidence_dict)

        for relation, evidences in relation_evidences.items():
            edges_entry.append(
                {
                    "source": u.md5,
                    "target": v.md5,
                    "relation": relation,
                    "label": u_v_r_bel[u, v, relation],
                    "metadata": {
                        "evidences": evidences,
                    },
                }
            )

    return {
        "graph": {
            "metadata": dict(
                origin=dict(name="pybel", version=get_version()),
                **graph.document,
            ),
            "nodes": nodes_entry,
            "edges": edges_entry,
        },
    }


@open_file(1, mode="w")
def to_jgif_file(graph: BELGraph, file: Union[str, TextIO], **kwargs) -> None:
    """Write JGIF to a file.

    :param graph: A BEL graph
    :param file: A writable file or file-like

    The example below shows how to output a BEL graph as JGIF to an open file.

    .. code-block:: python

       from pybel.examples import sialic_acid_graph
       from pybel import to_jgif_file
       with open('graph.bel.jgif.json', 'w') as file:
           to_jgif_file(sialic_acid_graph, file)

    The example below shows how to output a BEL graph as JGIF to a file at a given path.

    .. code-block:: python

        from pybel.examples import sialic_acid_graph
        from pybel import to_jgif_file
        to_jgif_file(sialic_acid_graph, 'graph.bel.jgif.json')

    If you have a big graph, you might consider storing it as a gzipped JGIF file
    by using :func:`to_jgif_gz`.
    """
    json.dump(to_jgif(graph), file, ensure_ascii=False, **kwargs)


def to_jgif_gz(graph, path: str, **kwargs) -> None:
    """Write a graph as JGIF JSON to a gzip file."""
    with gzip.open(path, "wt") as file:
        json.dump(to_jgif(graph), file, ensure_ascii=False, **kwargs)


def to_jgif_jsons(graph: BELGraph, **kwargs) -> str:
    """Dump this graph as a JGIF JSON object to a string."""
    return json.dumps(to_jgif(graph), ensure_ascii=False, **kwargs)


def post_jgif(graph: BELGraph, url: str, **kwargs) -> requests.Response:
    """Post the JGIF to a given URL."""
    return requests.post(url, json=to_jgif(graph), **kwargs)



# -*- coding: utf-8 -*-

"""Tests for the conversion procedure."""

import unittest
from typing import Tuple, Type

from pybel import BELGraph
from pybel.constants import (
    ASSOCIATION,
    DECREASES,
    DIRECTLY_DECREASES,
    DIRECTLY_INCREASES,
    EQUIVALENT_TO,
    INCREASES,
    IS_A,
    NEGATIVE_CORRELATION,
    PART_OF,
    POSITIVE_CORRELATION,
    REGULATES,
    RELATION,
    TARGET_MODIFIER,
)
from pybel.dsl import (
    Abundance,
    BaseEntity,
    BiologicalProcess,
    ComplexAbundance,
    MicroRna,
    NamedComplexAbundance,
    Pathology,
    Population,
    Protein,
    Rna,
    activity,
)
from pybel.io.triples import converters as tsvc
from pybel.io.triples.api import to_triple
from pybel.io.triples.converters import (
    AbundanceDirectlyDecreasesProteinActivityConverter,
    AbundanceDirectlyIncreasesProteinActivityConverter,
    AbundancePartOfPopulationConverter,
    AssociationConverter,
    Converter,
    CorrelationConverter,
    DecreasesAmountConverter,
    DrugIndicationConverter,
    DrugSideEffectConverter,
    EquivalenceConverter,
    IncreasesAmountConverter,
    IsAConverter,
    MiRNADecreasesExpressionConverter,
    PartOfNamedComplexConverter,
    RegulatesActivityConverter,
    RegulatesAmountConverter,
    SubprocessPartOfBiologicalProcessConverter,
    TranscriptionFactorForConverter,
)
from pybel.testing.utils import n
from pybel.typing import EdgeData


def _rel(x):
    return {RELATION: x}


def _rela(x, y=None):
    return {RELATION: x, TARGET_MODIFIER: activity(y)}


def _assoc(y):
    return {RELATION: ASSOCIATION, "association_type": y}


a1 = Abundance("CHEBI", "1")
p1 = Protein("HGNC", "1")
pf1 = Protein("INTERPRO", "1")
d1 = Pathology("MESH", "1")
b1 = BiologicalProcess("GO", "1")
b2 = BiologicalProcess("GO", "2")
m1 = MicroRna("MIRBASE", "1")
r1 = Rna("HGNC", "1")
r2 = Rna("HGNC", "2")
nca1 = NamedComplexAbundance("FPLX", "1")
pop1 = Population("taxonomy", "1")

p2 = Protein("HGNC", identifier="9236")
p3 = Protein("HGNC", identifier="9212")
r3 = p3.get_rna()
g3 = r3.get_gene()

c1 = ComplexAbundance([p2, g3])
c2 = ComplexAbundance([p1, p2])
c3 = ComplexAbundance([a1, p2])
p1_homodimer = ComplexAbundance([p1, p1])
p1_homotrimer = ComplexAbundance([p1, p1, p1])

converters_true_list = [
    (
        PartOfNamedComplexConverter,
        p1,
        nca1,
        _rel(PART_OF),
        ("HGNC:1", "partOf", "FPLX:1"),
    ),
    (
        SubprocessPartOfBiologicalProcessConverter,
        b1,
        b2,
        _rel(PART_OF),
        ("GO:1", "partOf", "GO:2"),
    ),
    (tsvc.ProcessCausalConverter, b1, b2, _rel(INCREASES), ("GO:1", INCREASES, "GO:2")),
    (tsvc.ProcessCausalConverter, b1, b2, _rel(DECREASES), ("GO:1", DECREASES, "GO:2")),
    (
        tsvc.ProcessCausalConverter,
        b1,
        b2,
        _rel(DIRECTLY_INCREASES),
        ("GO:1", DIRECTLY_INCREASES, "GO:2"),
    ),
    (
        tsvc.ProcessCausalConverter,
        b1,
        b2,
        _rel(DIRECTLY_DECREASES),
        ("GO:1", DIRECTLY_DECREASES, "GO:2"),
    ),
    (
        AssociationConverter,
        r1,
        r2,
        _rel(ASSOCIATION),
        ("HGNC:1", "association", "HGNC:2"),
    ),
    (
        AssociationConverter,
        r1,
        r2,
        _assoc("similarity"),
        ("HGNC:1", "similarity", "HGNC:2"),
    ),
    (
        CorrelationConverter,
        r1,
        r2,
        _rel(POSITIVE_CORRELATION),
        ("HGNC:1", "positiveCorrelation", "HGNC:2"),
    ),
    (IsAConverter, p1, pf1, _rel(IS_A), ("HGNC:1", "isA", "INTERPRO:1")),
    # Found in ADEPTUS
    (
        CorrelationConverter,
        d1,
        r1,
        _rel(POSITIVE_CORRELATION),
        ("MESH:1", "positiveCorrelation", "HGNC:1"),
    ),
    (
        CorrelationConverter,
        d1,
        r1,
        _rel(NEGATIVE_CORRELATION),
        ("MESH:1", "negativeCorrelation", "HGNC:1"),
    ),
    # Found in LINCS (not integrated yet)
    (
        RegulatesAmountConverter,
        a1,
        r1,
        _rel(REGULATES),
        ("CHEBI:1", "regulatesAmountOf", "HGNC:1"),
    ),
    (
        IncreasesAmountConverter,
        a1,
        r1,
        _rel(INCREASES),
        ("CHEBI:1", "increasesAmountOf", "HGNC:1"),
    ),
    (
        DecreasesAmountConverter,
        a1,
        r1,
        _rel(DECREASES),
        ("CHEBI:1", "decreasesAmountOf", "HGNC:1"),
    ),
    # Found in SIDER
    (
        DrugSideEffectConverter,
        a1,
        d1,
        _rel(INCREASES),
        ("CHEBI:1", "increases", "MESH:1"),
    ),
    (
        DrugIndicationConverter,
        a1,
        d1,
        _rel(DECREASES),
        ("CHEBI:1", "decreases", "MESH:1"),
    ),
    # Found in miRTarBase
    (
        MiRNADecreasesExpressionConverter,
        m1,
        r1,
        _rel(DECREASES),
        ("MIRBASE:1", "repressesExpressionOf", "HGNC:1"),
    ),
    # Found in chemogenomics databases (e.g., DrugBank)
    (
        RegulatesActivityConverter,
        a1,
        p1,
        _rela(REGULATES),
        ("CHEBI:1", "activityDirectlyRegulatesActivityOf", "HGNC:1"),
    ),
    (
        AbundanceDirectlyDecreasesProteinActivityConverter,
        a1,
        p1,
        _rela(DIRECTLY_DECREASES),
        ("CHEBI:1", "activityDirectlyNegativelyRegulatesActivityOf", "HGNC:1"),
    ),
    (
        AbundanceDirectlyIncreasesProteinActivityConverter,
        a1,
        p1,
        _rela(DIRECTLY_INCREASES),
        ("CHEBI:1", "activityDirectlyPositivelyRegulatesActivityOf", "HGNC:1"),
    ),
    # Found in ComPath
    (
        EquivalenceConverter,
        b1,
        b2,
        _rel(EQUIVALENT_TO),
        ("GO:1", "equivalentTo", "GO:2"),
    ),
    (
        SubprocessPartOfBiologicalProcessConverter,
        b1,
        b2,
        _rel(PART_OF),
        ("GO:1", "partOf", "GO:2"),
    ),
    # Misc
    (
        AbundancePartOfPopulationConverter,
        a1,
        pop1,
        _rel(PART_OF),
        ("CHEBI:1", "partOf", "taxonomy:1"),
    ),
    # complex(g(hgnc:9212), p(hgnc:9236)) directlyIncreases r(hgnc:9212)
    (
        TranscriptionFactorForConverter,
        c1,
        r3,
        _rel(DIRECTLY_INCREASES),
        ("HGNC:9236", DIRECTLY_INCREASES, "HGNC:9212"),
    ),
    (
        TranscriptionFactorForConverter,
        c1,
        r3,
        _rel(DIRECTLY_DECREASES),
        ("HGNC:9236", DIRECTLY_DECREASES, "HGNC:9212"),
    ),
    (
        tsvc.BindsGeneConverter,
        p2,
        c1,
        _rel(DIRECTLY_INCREASES),
        (p2.curie, "bindsToGene", g3.curie),
    ),
    (
        tsvc.BindsProteinConverter,
        p1,
        c2,
        _rel(DIRECTLY_INCREASES),
        (p1.curie, "bindsToProtein", p2.curie),
    ),
    (
        tsvc.BindsProteinConverter,
        a1,
        c3,
        _rel(DIRECTLY_INCREASES),
        (a1.curie, "bindsToProtein", p2.curie),
    ),
    (
        tsvc.HomomultimerConverter,
        p1,
        p1_homodimer,
        _rel(DIRECTLY_INCREASES),
        (p1.curie, "bindsToProtein", p1.curie),
    ),
    (
        tsvc.HomomultimerConverter,
        p1,
        p1_homotrimer,
        _rel(DIRECTLY_INCREASES),
        (p1.curie, "bindsToProtein", p1.curie),
    ),
    (
        tsvc.ProteinRegulatesComplex,
        p3,
        c2,
        _rel(DIRECTLY_INCREASES),
        (p3.curie, "increasesAmountOf", str(c2)),
    ),
    (
        tsvc.ProteinRegulatesComplex,
        p3,
        c2,
        _rel(DIRECTLY_DECREASES),
        (p3.curie, "decreasesAmountOf", str(c2)),
    ),
    (
        tsvc.ProteinRegulatesComplex,
        p3,
        c2,
        _rel(DECREASES),
        (p3.curie, "decreasesAmountOf", str(c2)),
    ),
    (
        tsvc.ProteinRegulatesComplex,
        p3,
        c2,
        _rel(INCREASES),
        (p3.curie, "increasesAmountOf", str(c2)),
    ),
]

converters_false_list = [
    (PartOfNamedComplexConverter, nca1, p1, _rel(IS_A)),
]


class TestConverters(unittest.TestCase):
    """Tests for the converter classes."""

    def help_test_convert(
        self,
        converter: Type[Converter],
        u: BaseEntity,
        v: BaseEntity,
        edge_data: EdgeData,
        triple: Tuple[str, str, str],
    ) -> None:
        """Test a converter class."""
        self.assertTrue(
            issubclass(converter, Converter),
            msg="Not a Converter: {}".format(converter.__name__),
        )
        key = n()
        self.assertTrue(
            converter.predicate(u, v, key, edge_data),
            msg="Predicate failed: {}".format(converter.__name__),
        )
        self.assertEqual(
            triple,
            converter.convert(u, v, key, edge_data),
            msg="Conversion failed: {}".format(converter.__name__),
        )
        graph = BELGraph()
        graph.add_edge(u, v, key=key, **edge_data)
        self.assertEqual(
            triple,
            to_triple(graph, u, v, key),
            msg="get_triple failed: {}".format(converter.__name__),
        )

    def test_converters_true(self):
        """Test passing converters."""
        for converter, u, v, edge_data, triple in converters_true_list:
            with self.subTest(converter=converter.__qualname__):
                self.help_test_convert(converter, u, v, edge_data, triple)

    def test_converters_false(self):
        """Test falsification of converters."""
        for converter, u, v, edge_data in converters_false_list:
            with self.subTest(converter=converter.__qualname__):
                self.assertFalse(converter.predicate(u, v, n(), edge_data))



# -*- coding: utf-8 -*-

"""Resources that don't change over time for PyBEL."""

SPECIES_PATTERN = r"^\d+$"
CONFIDENCE_URL = "https://arty.scai.fraunhofer.de/artifactory/bel/annotation/confidence/confidence-1.0.0.belanno"



# -*- coding: utf-8 -*-

"""Mutations that expand the graph."""

from . import neighborhood, upstream
from .neighborhood import *
from .upstream import *

__all__ = [k for k in locals() if not k.startswith("_")]
